import numpy as np

# the parameters ranges are partially derived from https://github.com/EpistasisLab/tpot
params_range_by_model = {
    'kmeans': {'n_clusters': list(range(2, 5))},
    'knn': {
        'n_neighbors': list(range(1, 50)),
        'weights': ["uniform", "distance"],
        'p': [1, 2]},
    'logit': {
        'C': [1e-2, 1e-1, 0.5, 0.9, 1., 2., 5., 10.]},
    'xgboost': {
        'n_estimators': [100],
        'max_depth': list(range(1, 7)),
        'learning_rate': list(np.arange(0.1, 0.9, 0.1)),
        'subsample': list(np.arange(0.05, 1.01, 0.05)),
        'min_child_weight': list(range(1, 21)),
        'nthread': [1]},
    'rf': {
        'n_estimators': [100],
        'criterion': ["gini", "entropy"],
        'max_features': np.arange(0.05, 1.01, 0.05),
        'min_samples_split': list(range(2, 10)),
        'min_samples_leaf': list(range(1, 15)),
        'bootstrap': [True, False]},
    'xgbreg': {
        'n_estimators': [100],
        'max_depth': range(1, 11),
        'learning_rate': [1e-3, 1e-2, 1e-1, 0.5, 1.],
        'subsample': np.arange(0.05, 1.01, 0.05),
        'min_child_weight': range(1, 21),
        'objective': ['reg:squarederror']
    },
    'rfr': {
        'n_estimators': [100],
        'max_features': np.arange(0.05, 1.01, 0.05),
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21),
        'bootstrap': [True, False]
    },
    'svr': {
        'loss': ["epsilon_insensitive", "squared_epsilon_insensitive"],
        'dual': [True, False],
        'tol': [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
        'C': [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1., 5., 10., 15., 20., 25.],
        'epsilon': [1e-4, 1e-3, 1e-2, 1e-1, 1.]
    },
    'knnreg': {
        'n_neighbors': range(1, 101),
        'weights': ["uniform", "distance"],
        'p': [1, 2]
    },
    'dtreg': {
        'max_depth': range(1, 11),
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21)
    },
    'adareg': {
        'n_estimators': [100],
        'learning_rate': [1e-3, 1e-2, 1e-1, 0.5, 1.],
        'loss': ["linear", "square", "exponential"]
    },
    'gbr': {
        'n_estimators': [100],
        'loss': ["ls", "lad", "huber", "quantile"],
        'learning_rate': [1e-3, 1e-2, 1e-1, 0.5, 1.],
        'max_depth': range(1, 11),
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21),
        'subsample': np.arange(0.05, 1.01, 0.05),
        'max_features': np.arange(0.05, 1.01, 0.05),
        'alpha': [0.75, 0.8, 0.85, 0.9, 0.95, 0.99]
    },
    'treg': {
        'n_estimators': [100],
        'max_features': np.arange(0.05, 1.01, 0.05),
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21),
        'bootstrap': [True, False]
    }
}



from copy import deepcopy
from typing import (Any, List)

from fedot.core.composer.optimisers.selection import SelectionTypesEnum, individuals_selection
from fedot.core.utils import ComparableEnum as Enum
from functools import partial


class GeneticSchemeTypesEnum(Enum):
    steady_state = 'steady_state'
    generational = 'generational'
    parameter_free = 'parameter_free'


def inheritance(type: GeneticSchemeTypesEnum, selection_types: List[SelectionTypesEnum],
                prev_population: List[Any], new_population: List[Any], max_size: int) -> List[Any]:
    steady_state_scheme = partial(steady_state_inheritance, selection_types, prev_population, new_population, max_size)
    generational_scheme = partial(direct_inheritance, new_population, max_size)
    inheritance_type_by_genetic_scheme = {
        GeneticSchemeTypesEnum.generational: generational_scheme,
        GeneticSchemeTypesEnum.steady_state: steady_state_scheme,
        GeneticSchemeTypesEnum.parameter_free: steady_state_scheme
    }
    return inheritance_type_by_genetic_scheme[type]()


def steady_state_inheritance(selection_types: List[SelectionTypesEnum],
                             prev_population: List[Any],
                             new_population: List[Any], max_size: int):
    return individuals_selection(types=selection_types, individuals=prev_population + new_population, pop_size=max_size)


def direct_inheritance(new_population: List[Any], max_size: int):
    return deepcopy(new_population[:max_size])



import json
import os
import shutil

import pytest

from cases.data.data_utils import get_scoring_case_data_paths
from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import InputData
from fedot.utilities.synthetic.chain_template_new import ChainTemplate, extract_subtree_root


@pytest.fixture(scope="session", autouse=True)
def creation_model_files_before_after_tests(request):
    create_json_models_files()
    request.addfinalizer(delete_json_models_files)


def create_json_models_files():
    """
    Creating JSON's files for test.
    """
    chain = create_chain()
    chain.save_chain("data/test_chain_convert_to_json.json")

    chain_fitted = create_fitted_chain()
    chain_fitted.save_chain("data/test_fitted_chain_convert_to_json.json")

    chain_empty = Chain()
    chain_empty.save_chain("data/test_empty_chain_convert_to_json.json")


def delete_json_models_files():
    """
    Delete JSON's files.
    """
    with open("data/test_fitted_chain_convert_to_json.json", 'r') as json_file:
        chain_fitted_object = json.load(json_file)

    delete_fitted_models(chain_fitted_object)

    os.remove("data/test_fitted_chain_convert_to_json.json")
    os.remove("data/test_empty_chain_convert_to_json.json")
    os.remove("data/test_chain_convert_to_json.json")


def delete_fitted_models(chain):
    """
    Delete directory and chain's local fitted models.

    :param chain: chain which model's need to delete
    """
    model_path = chain['nodes'][0]['trained_model_path']
    dir_path = os.path.dirname(os.path.abspath(model_path))
    shutil.rmtree(dir_path)


def create_chain() -> Chain:
    chain = Chain()
    node_logit = PrimaryNode('logit')

    node_lda = PrimaryNode('lda')
    node_lda.custom_params = {'n_components': 1}

    node_xgboost = PrimaryNode('xgboost')

    node_knn = PrimaryNode('knn')
    node_knn.custom_params = {'n_neighbors': 9}

    node_knn_second = SecondaryNode('knn')
    node_knn_second.custom_params = {'n_neighbors': 5}
    node_knn_second.nodes_from = [node_lda, node_knn]

    node_logit_second = SecondaryNode('logit')
    node_logit_second.nodes_from = [node_xgboost, node_lda]

    node_lda_second = SecondaryNode('lda')
    node_lda_second.custom_params = {'n_components': 1}
    node_lda_second.nodes_from = [node_logit_second, node_knn_second, node_logit]

    node_xgboost_second = SecondaryNode('xgboost')
    node_xgboost_second.nodes_from = [node_logit, node_logit_second, node_knn]

    node_knn_third = SecondaryNode('knn')
    node_knn_third.custom_params = {'n_neighbors': 8}
    node_knn_third.nodes_from = [node_lda_second, node_xgboost_second]

    chain.add_node(node_knn_third)

    return chain


def create_fitted_chain() -> Chain:
    train_file_path, test_file_path = get_scoring_case_data_paths()
    train_data = InputData.from_csv(train_file_path)

    chain = create_chain()
    chain.fit(train_data)

    return chain


def test_export_chain_to_json_correctly():
    chain = create_chain()
    json_actual = chain.save_chain("data/1.json")

    with open("data/test_chain_convert_to_json.json", 'r') as json_file:
        json_expected = json.load(json_file)

    os.remove("data/1.json")
    assert json_actual == json.dumps(json_expected)


def test_chain_template_to_json_correctly():
    chain = create_chain()
    chain_template = ChainTemplate(chain)
    json_actual = chain_template.convert_to_dict()

    with open("data/test_chain_convert_to_json.json", 'r') as json_file:
        json_expected = json.load(json_file)

    assert json.dumps(json_actual) == json.dumps(json_expected)


def test_import_json_to_chain_correctly():
    chain = Chain()
    chain.load_chain("data/test_chain_convert_to_json.json")
    json_actual = chain.save_chain("data/1.json")

    chain_expected = create_chain()
    json_expected = chain_expected.save_chain("data/2.json")

    os.remove("data/1.json")
    os.remove("data/2.json")
    assert json.dumps(json_actual) == json.dumps(json_expected)


def test_import_json_template_to_chain_correctly():
    chain = Chain()
    chain_template = ChainTemplate(chain)
    chain_template.import_from_json("data/test_chain_convert_to_json.json")
    json_actual = chain_template.convert_to_dict()

    chain_expected = create_chain()
    chain_expected_template = ChainTemplate(chain_expected)
    json_expected = chain_expected_template.convert_to_dict()

    assert json.dumps(json_actual) == json.dumps(json_expected)


def test_import_json_to_fitted_chain_correctly():
    chain = Chain()
    chain.load_chain("data/test_fitted_chain_convert_to_json.json")
    json_actual = chain.save_chain("data/1.json")

    with open("data/test_fitted_chain_convert_to_json.json", 'r') as json_file:
        json_expected = json.load(json_file)

    os.remove("data/1.json")
    assert json_actual == json.dumps(json_expected)


def test_import_json_to_fitted_chain_template_correctly():
    chain = Chain()
    chain_template = ChainTemplate(chain)
    chain_template.import_from_json("data/test_fitted_chain_convert_to_json.json")
    json_actual = chain_template.convert_to_dict()

    with open("data/test_fitted_chain_convert_to_json.json", 'r') as json_file:
        json_expected = json.load(json_file)

    assert json.dumps(json_actual) == json.dumps(json_expected)


def test_empty_chain_to_json_correctly():
    chain = Chain()
    chain_template = ChainTemplate(chain)
    json_actual = chain_template.convert_to_dict()

    with open("data/test_empty_chain_convert_to_json.json", 'r') as json_file:
        json_expected = json.load(json_file)

    assert json.dumps(json_actual) == json.dumps(json_expected)


def test_export_import_for_one_chain_object_correctly():
    """
    This test checks whether it is possible to upload a new model to the same object. In other words,
     apply a sequence of commands to the chain object:
    - load_chain(path_first)
    - load_chain(path_second)
    - load_chain(path_third)
    and the last command will rewrite the chain object correctly.
    """
    chain_fitted = create_fitted_chain()
    json_first = chain_fitted.save_chain("data/2.json")

    chain_fitted_after = create_chain()
    chain_fitted_after.save_chain("data/1.json")
    chain_fitted_after.load_chain("data/2.json")

    json_second = chain_fitted_after.save_chain("data/3.json")

    for i in range(1, 4):
        os.remove(f"data/{i}.json")

    delete_fitted_models(json.loads(json_first))
    assert json_first == json_second


def test_absolute_relative_paths_correctly_no_exception():
    chain = create_chain()
    chain.save_chain("data/test/1.json")

    absolute_path = os.path.join(os.path.abspath("data/2.json"))
    chain.save_chain(absolute_path)

    chain.load_chain("data/test/1.json")
    chain.load_chain(absolute_path)

    os.remove("data/test/1.json")
    os.remove(absolute_path)
    os.rmdir("data/test")


def test_import_custom_json_object_to_chain_and_fit_correctly_no_exception():
    train_file_path, test_file_path = get_scoring_case_data_paths()
    train_data = InputData.from_csv(train_file_path)

    data_path = str(os.path.dirname(__file__))
    json_file_path = os.path.join(data_path, '..', 'test', 'data', 'test_custom_json_template.json')

    chain = Chain()
    chain_template = ChainTemplate(chain)
    chain_template.import_from_json(json_file_path)

    chain.fit(train_data)
    json_actual = chain.save_chain("data/1.json")

    delete_fitted_models(json.loads(json_actual))
    os.remove("data/1.json")


def create_four_depth_chain():
    knn_node = PrimaryNode('knn')
    lda_node = PrimaryNode('lda')
    xgb_node = PrimaryNode('xgboost')
    logit_node = PrimaryNode('logit')

    logit_node_second = SecondaryNode('logit', nodes_from=[knn_node, lda_node])
    xgb_node_second = SecondaryNode('xgboost', nodes_from=[logit_node])

    qda_node_third = SecondaryNode('qda', nodes_from=[xgb_node_second])
    knn_node_third = SecondaryNode('knn', nodes_from=[logit_node_second, xgb_node])

    knn_root = SecondaryNode('knn', nodes_from=[qda_node_third, knn_node_third])

    chain = Chain()
    chain.add_node(knn_root)

    return chain


def test_extract_subtree_root():
    chain = create_four_depth_chain()
    chain_template = ChainTemplate(chain)

    expected_types = ['knn', 'logit', 'knn', 'lda', 'xgboost']
    new_root_node_id = 4

    root_node = extract_subtree_root(root_model_id=new_root_node_id,
                                     chain_template=chain_template)

    sub_chain = Chain()
    sub_chain.add_node(root_node)
    actual_types = [node.model.model_type for node in sub_chain.nodes]

    assertion_list = [True if expected_types[index] == actual_types[index] else False
                      for index in range(len(expected_types))]
    assert all(assertion_list)



import datetime

from fedot.core.composer.gp_composer.gp_composer import GPComposerRequirements, GPComposerBuilder
from fedot.core.composer.optimisers.gp_optimiser import GPChainOptimiserParameters, GeneticSchemeTypesEnum
from fedot.core.repository.model_types_repository import ModelTypesRepository
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum, MetricsRepository
from fedot.core.repository.tasks import Task, TaskTypesEnum


def test_gp_composer_builder():
    task = Task(TaskTypesEnum.classification)

    available_model_types, _ = ModelTypesRepository().suitable_model(task_type=task.task_type)

    metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC)

    composer_requirements = GPComposerRequirements(
        primary=available_model_types,
        secondary=available_model_types, max_arity=3,
        max_depth=3, pop_size=5, num_of_generations=4,
        crossover_prob=0.8, mutation_prob=1, max_lead_time=datetime.timedelta(minutes=5))

    scheme_type = GeneticSchemeTypesEnum.steady_state
    optimiser_parameters = GPChainOptimiserParameters(genetic_scheme_type=scheme_type)

    builder_with_custom_params = GPComposerBuilder(task=task).with_requirements(composer_requirements).with_metrics(
        metric_function).with_optimiser_parameters(optimiser_parameters)

    composer_with_custom_params = builder_with_custom_params.build()

    assert composer_with_custom_params.optimiser.parameters.genetic_scheme_type == scheme_type
    assert composer_with_custom_params.metrics == metric_function
    assert composer_with_custom_params.composer_requirements.pop_size == 5
    assert composer_with_custom_params.composer_requirements.mutation_prob == 1

    builder_with_default_params = GPComposerBuilder(task=task)
    composer_with_default_params = builder_with_default_params.build()

    default_metric = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC.ROCAUC_penalty)

    assert composer_with_default_params.optimiser.parameters.genetic_scheme_type == GeneticSchemeTypesEnum.generational
    assert composer_with_default_params.metrics == default_metric
    assert composer_with_default_params.composer_requirements.pop_size == 20
    assert composer_with_default_params.composer_requirements.mutation_prob == 0.8



from copy import copy
from typing import List

import numpy as np


def prepare_lagged_ts_for_prediction(data: 'InputData', is_for_fit: bool = True):
    criteria = ['features']

    # nan in target is not acceptable in fit step
    if is_for_fit:
        criteria.append('target')
        if data.features is not None:
            data.features = data.features[0:len(data.target), ...]

    cleaned_data = _clean_nans_in_lagged_features(data, criteria)

    return cleaned_data


def _clean_nans_in_lagged_features(data: 'InputData', criteria: List[str]):
    """Removes data items that corresponds to NaNs in array_with_nans
    :param data: dataset to filter
    :param criteria: list of filtering criterion names (possible variants are features, target, idx)
    :return: dataset without NaNs
    """

    data_to_clean = copy(data)

    for criterion in criteria:

        datasets_for_criterion = {
            'features': data_to_clean.features,
            'target': data_to_clean.target,
            'idx': data_to_clean.idx
        }
        array_with_nans = datasets_for_criterion[criterion]

        if array_with_nans is not None:
            nans = np.isnan(array_with_nans)
            idx = data_to_clean.idx
            features = data_to_clean.features
            target = data_to_clean.target

            # TODO try to simplify the 1D/2D logic
            # remove all rows with nan in array_with_nans
            if len(array_with_nans.shape) == 1:
                if not nans.any():
                    # if there is no nans at all, go to next criterion
                    continue

                idx = idx[~nans[0:len(idx)]]
                features = features[~nans[0:len(features)]] if data.features is not None else features
                target = target[~nans[0:len(target)]] if data.target is not None else target
            elif len(array_with_nans.shape) == 2:
                if not nans.any().any():
                    # if there is no nans at all, go to next criterion
                    continue
                idx = idx[~nans.any(axis=1)[0:len(idx)]]
                features = features[~nans.any(axis=1)[0:len(features)]] if features is not None else features
                target = target[~nans.any(axis=1)[0:len(target)]] if target is not None else target
            else:
                raise NotImplementedError('Dimensionality not supported')

            data_to_clean.idx = idx
            data_to_clean.features = features
            data_to_clean.target = target

    return data_to_clean



from copy import deepcopy
from functools import partial
from random import choice, randint, random
from typing import Any

from fedot.core.composer.chain import Chain, List
from fedot.core.composer.constraint import constraint_function
from fedot.core.composer.optimisers.gp_operators import node_depth, nodes_from_height, random_chain, node_height
from fedot.core.utils import ComparableEnum as Enum


class MutationTypesEnum(Enum):
    simple = 'simple'
    growth = 'growth'
    local_growth = 'local_growth'
    reduce = 'reduce'
    none = 'none'


class MutationStrengthEnum(Enum):
    weak = 0.2
    mean = 1.0
    strong = 5.0


def get_mutation_prob(mut_id, root_node):
    default_mutation_prob = 0.7
    if mut_id in list(MutationStrengthEnum):
        mutation_strength = mut_id.value
        mutation_prob = mutation_strength / (node_depth(root_node) + 1)
    else:
        mutation_prob = default_mutation_prob
    return mutation_prob


def mutation(types: List[MutationTypesEnum], chain_generation_params, chain: Chain, requirements,
             max_depth: int = None) -> Any:
    max_depth = max_depth if max_depth else requirements.max_depth
    mutation_prob = requirements.mutation_prob
    if mutation_prob and random() > mutation_prob:
        return deepcopy(chain)

    type = choice(types)
    if type == MutationTypesEnum.none:
        new_chain = deepcopy(chain)
    elif type in mutation_by_type:
        is_correct_chain = False
        while not is_correct_chain:
            if type in (MutationTypesEnum.growth, MutationTypesEnum.local_growth):
                new_chain = mutation_by_type[type](chain=deepcopy(chain), requirements=requirements,
                                                   chain_generation_params=chain_generation_params, max_depth=max_depth)
            else:
                new_chain = mutation_by_type[type](chain=deepcopy(chain), requirements=requirements,
                                                   chain_generation_params=chain_generation_params)
            is_correct_chain = constraint_function(new_chain)
    else:
        raise ValueError(f'Required mutation type is not found: {type}')

    return new_chain


def simple_mutation(chain: Any, requirements, chain_generation_params) -> Any:
    """
    This type of mutation is passed over all nodes of the tree started from the root node and changes
    nodes’ models with probability - 'node mutation probability' which is inicialised inside the function
    """

    node_mutation_probability = get_mutation_prob(mut_id=requirements.mutation_strength,
                                                  root_node=chain.root_node)

    def replace_node_to_random_recursive(node: Any) -> Any:
        if node.nodes_from:
            if random() < node_mutation_probability:
                secondary_node = chain_generation_params.secondary_node_func(model_type=choice(requirements.secondary),
                                                                             nodes_from=node.nodes_from)
                chain.update_node(node, secondary_node)
            for child in node.nodes_from:
                replace_node_to_random_recursive(child)
        else:
            if random() < node_mutation_probability:
                primary_node = chain_generation_params.primary_node_func(model_type=choice(requirements.primary))
                chain.update_node(node, primary_node)

    replace_node_to_random_recursive(chain.root_node)

    return chain


def growth_mutation(chain: Any, requirements, chain_generation_params, max_depth: int, local_growth=True) -> Any:
    """
    This mutation selects a random node in a tree, generates new subtree, and replaces the selected node's subtree.
    :param local_growth: if true then maximal depth of new subtree equals depth of tree located in
    selected random node, if false then previous depth of selected node doesn't affect to new subtree depth,
    maximal depth of new subtree just should satisfy depth constraint in parent tree
    """

    random_layer_in_chain = randint(0, chain.depth - 1)
    node_from_chain = choice(nodes_from_height(chain, random_layer_in_chain))
    if local_growth:
        is_primary_node_selected = (not node_from_chain.nodes_from) or (
                node_from_chain.nodes_from and node_from_chain != chain.root_node and randint(0, 1))
    else:
        is_primary_node_selected = randint(0, 1) and not node_height(chain, node_from_chain) \
                                                         < max_depth
    if is_primary_node_selected:
        new_subtree = chain_generation_params.primary_node_func(model_type=choice(requirements.primary))
    else:
        if local_growth:
            max_depth = node_depth(node_from_chain)
        else:
            max_depth = max_depth - node_height(chain, node_from_chain)
        new_subtree = random_chain(chain_generation_params=chain_generation_params, requirements=requirements,
                                   max_depth=max_depth).root_node
    chain.replace_node_with_parents(node_from_chain, new_subtree)
    return chain


def reduce_mutation(chain: Any, requirements, chain_generation_params) -> Any:
    """
    Selects a random node in a tree, then removes its subtree. If the current arity of the node's
    parent is more than the specified minimal arity, then the selected node is also removed.
    Otherwise, it is replaced by a random primary node.
    """

    nodes = [node for node in chain.nodes if node is not chain.root_node]
    node_to_del = choice(nodes)
    childs = chain.node_childs(node_to_del)
    is_possible_to_delete = all([len(child.nodes_from) - 1 >= requirements.min_arity for child in childs])
    if is_possible_to_delete:
        chain.delete_node(node_to_del)
    else:
        primary_node = chain_generation_params.primary_node_func(model_type=choice(requirements.primary))
        chain.replace_node_with_parents(node_to_del, primary_node)
    return chain


mutation_by_type = {
    MutationTypesEnum.simple: simple_mutation,
    MutationTypesEnum.growth: partial(growth_mutation, local_growth=False),
    MutationTypesEnum.local_growth: partial(growth_mutation, local_growth=True),
    MutationTypesEnum.reduce: reduce_mutation
}



import itertools
from dataclasses import dataclass
from functools import partial
from sys import maxsize as max_int_value
from typing import (
    Callable,
    Optional
)

from fedot.core.chain_validation import validate
from fedot.core.composer.chain import Chain, SharedChain
from fedot.core.composer.composer import Composer, ComposerRequirements
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.composer.optimisers.gp_optimiser import GPChainOptimiser, GPChainOptimiserParameters
from fedot.core.composer.optimisers.param_free_gp_optimiser import GPChainParameterFreeOptimiser
from fedot.core.composer.optimisers.inheritance import GeneticSchemeTypesEnum
from fedot.core.composer.optimisers.mutation import MutationStrengthEnum
from fedot.core.composer.visualisation import ComposerVisualiser
from fedot.core.composer.write_history import write_composer_history_to_csv
from fedot.core.models.data import InputData, train_test_data_setup
from fedot.core.repository.model_types_repository import ModelTypesRepository
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum, RegressionMetricsEnum, \
    MetricsRepository
from fedot.core.repository.tasks import Task, TaskTypesEnum


@dataclass
class GPComposerRequirements(ComposerRequirements):
    """
    Dataclass is for defining the requirements for composition process of genetic programming composer

    :param pop_size: population size
    :param num_of_generations: maximal number of evolutionary algorithm generations
    :param crossover_prob: crossover probability (the chance that two chromosomes exchange some of their parts)
    :param mutation_prob: mutation probability
    :param mutation_strength: strength of mutation in tree (using in certain mutation types)
    """
    pop_size: Optional[int] = 20
    num_of_generations: Optional[int] = 100
    crossover_prob: Optional[float] = 0.8
    mutation_prob: Optional[float] = 0.8
    mutation_strength: MutationStrengthEnum = MutationStrengthEnum.mean


@dataclass
class ChainGenerationParams:
    """
    This dataclass is for defining the parameters using in chain generation process

    :param primary_node_func: the function for primary node generation
    :param secondary_node_func: the function for secondary node generation
    :param chain_class: class for the chain object
    """
    primary_node_func: Callable = PrimaryNode
    secondary_node_func: Callable = SecondaryNode
    chain_class: Callable = Chain


class GPComposer(Composer):
    """
    Genetic programming based composer
    :param optimiser: optimiser generated in GPComposerBuilder
    :param metrics: metrics used to define the quality of found solution
    :param composer_requirements: requirements for composition process
    :param initial_chain: defines the initial state of the population. If None then initial population is random.
    """

    def __init__(self, optimiser=None,
                 composer_requirements: Optional[GPComposerRequirements] = None,
                 metrics: Optional[Callable] = None,
                 initial_chain: Optional[Chain] = None):

        super().__init__(metrics=metrics, composer_requirements=composer_requirements, initial_chain=initial_chain)
        self.shared_cache = {}
        self.optimiser = optimiser

    def compose_chain(self, data: InputData, is_visualise: bool = False, is_tune: bool = False) -> Chain:

        if not self.optimiser:
            raise AttributeError(f'Optimiser for chain composition is not defined')

        train_data, test_data = train_test_data_setup(data, 0.8)
        self.shared_cache.clear()
        metric_function_for_nodes = partial(self.metric_for_nodes,
                                            self.metrics, train_data, test_data, True)

        best_chain, self.history = self.optimiser.optimise(metric_function_for_nodes)

        self.log.info('GP composition finished')

        if is_visualise:
            historical_fitness = [[chain.fitness for chain in pop] for pop in self.history]
            all_historical_fitness = list(itertools.chain(*historical_fitness))
            historical_chains = list(itertools.chain(*self.history))
            ComposerVisualiser.visualise_history(historical_chains, all_historical_fitness)

        write_composer_history_to_csv(historical_chains=self.history)

        if is_tune:
            self.tune_chain(best_chain, data, self.composer_requirements.max_lead_time)
        return best_chain

    def metric_for_nodes(self, metric_function, train_data: InputData,
                         test_data: InputData, is_chain_shared: bool,
                         chain: Chain) -> float:
        try:
            validate(chain)
            if is_chain_shared:
                chain = SharedChain(base_chain=chain, shared_cache=self.shared_cache)
            chain.fit(input_data=train_data)
            return metric_function(chain, test_data)
        except Exception as ex:
            self.log.info(f'Error in chain assessment during composition: {ex}. Continue.')
            return max_int_value

    @staticmethod
    def tune_chain(chain: Chain, data: InputData, time_limit):
        chain.fine_tune_all_nodes(input_data=data, max_lead_time=time_limit)


class GPComposerBuilder:
    def __init__(self, task: Task):
        self._composer = GPComposer()
        self.optimiser_parameters = GPChainOptimiserParameters()
        self.task = task
        self.set_default_composer_params()

    def with_optimiser_parameters(self, optimiser_parameters):
        self.optimiser_parameters = optimiser_parameters
        return self

    def with_requirements(self, requirements):
        self._composer.composer_requirements = requirements
        return self

    def with_metrics(self, metrics):
        self._composer.metrics = metrics
        return self

    def with_initial_chain(self, initial_chain):
        self._composer.initial_chain = initial_chain
        return self

    def set_default_composer_params(self):
        if not self._composer.composer_requirements:
            models, _ = ModelTypesRepository().suitable_model(task_type=self.task.task_type)
            self._composer.composer_requirements = GPComposerRequirements(primary=models, secondary=models)
        if not self._composer.metrics:
            metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)
            if self.task.task_type in (TaskTypesEnum.regression, TaskTypesEnum.ts_forecasting):
                metric_function = MetricsRepository().metric_by_id(RegressionMetricsEnum.RMSE)
            self._composer.metrics = metric_function

    def build(self) -> Composer:
        optimiser_type = GPChainOptimiser
        if self.optimiser_parameters.genetic_scheme_type == GeneticSchemeTypesEnum.parameter_free:
            optimiser_type = GPChainParameterFreeOptimiser

        chain_generation_params = ChainGenerationParams()

        optimiser = optimiser_type(initial_chain=self._composer.initial_chain,
                                   requirements=self._composer.composer_requirements,
                                   chain_generation_params=chain_generation_params,
                                   parameters=self.optimiser_parameters, log=self._composer.log)

        self._composer.optimiser = optimiser

        return self._composer



import csv
import os
from typing import (Any, List)

from fedot.core.utils import default_fedot_data_dir


def write_composer_history_to_csv(historical_chains: List[Any], file='history.csv'):
    history_dir = os.path.join(default_fedot_data_dir(), 'composing_history')
    file = f'{history_dir}/{file}'
    if not os.path.isdir(history_dir):
        os.mkdir(history_dir)
    write_header_to_csv(file)
    i = 0
    for gen_num, gen_chains in enumerate(historical_chains):
        for chain in gen_chains:
            add_history_to_csv(file, chain.fitness, len(chain.nodes), chain.depth, i, gen_num)
            i += 1


def write_header_to_csv(f):
    with open(f, 'w', newline='') as file:
        writer = csv.writer(file, quoting=csv.QUOTE_ALL)
        writer.writerow(['num', 'generation', 'fitness, num_of_models, depth'])


def add_history_to_csv(f, fitness: float, models_num: int, depth: int, num: int = None, generation: int = None):
    with open(f, 'a', newline='') as file:
        writer = csv.writer(file, quoting=csv.QUOTE_ALL)
        writer.writerow([num, generation, fitness, models_num, depth])



import numpy as np
from sklearn.metrics import mean_squared_error as mse_metric

from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TsForecastingParams


def multistep_prediction_to_ts(prediction):
    # choose the early steps of first prediction
    first_pred_part = prediction[0, :-1]
    # choose last forecasting step only for each prediction
    prediction = prediction[:, -1]
    return np.concatenate((first_pred_part, prediction))


def cut_future_prediction_part(prediction, task_params: TsForecastingParams):
    # TODO add multivariate

    length_of_cut = task_params.forecast_length
    return prediction[:-length_of_cut]


def preserve_prediction_length(prediction, expected_length: int):
    if len(prediction) < expected_length:
        zeros = expected_length - len(prediction)
        zeros = np.asarray([np.nan] * zeros)
        if len(prediction.shape) == 1:
            prediction = np.concatenate((zeros, prediction))
        else:
            prediction_steps = []
            for forecast_depth in range(prediction.shape[1]):
                prediction_steps.append(np.concatenate((zeros, prediction[:, forecast_depth])))
            prediction = np.stack(np.asarray(prediction_steps)).T
    return prediction


def post_process_forecasted_ts(prediction, input_data: 'InputData'):
    task = input_data.task
    expected_length = len(input_data.idx)

    data_type = input_data.data_type

    if data_type == DataTypesEnum.ts_lagged_table:
        # because one step is devoted to the feature vector for future prediction
        expected_length = expected_length - 1

    if not task.task_params.return_all_steps and len(prediction.shape) > 1:
        prediction = multistep_prediction_to_ts(prediction)

    if not task.task_params.make_future_prediction and \
            data_type in [DataTypesEnum.ts_lagged_table]:
        # cut unwanted out-of-sample prediction
        prediction = cut_future_prediction_part(prediction, task.task_params)

    prediction = preserve_prediction_length(prediction, expected_length)

    return prediction


def ts_mse(obs, pred) -> float:
    return mse_metric(y_true=obs[~np.isnan(pred)],
                      y_pred=pred[~np.isnan(pred)],
                      squared=False)



import numpy as np
from sklearn.multiclass import OneVsRestClassifier
from sklearn.svm import SVC


class CustomSVC:
    def __init__(self):
        self.fitted_model = None
        self.classes_ = None

    def fit(self, train_data: np.array, target_data: np.array):
        self.fitted_model = OneVsRestClassifier(SVC(kernel='linear',
                                                    probability=True,
                                                    class_weight='balanced'))
        self.classes_ = np.unique(target_data)
        self.fitted_model.fit(train_data, target_data)
        return self.fitted_model

    def predict(self, data_to_predict: np.array):
        return self.fitted_model.predict(data_to_predict)

    def predict_proba(self, data_to_predict: np.array):
        return self.fitted_model.predict_proba(data_to_predict)

    def get_params(self):
        return self.fitted_model.get_params()



import os

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.log import default_log
from examples.chain_tune import get_case_train_test_data


def get_simple_chain(log):
    first = PrimaryNode(model_type='xgboost', log=log)
    second = PrimaryNode(model_type='knn', log=log)
    final = SecondaryNode(model_type='logit',
                          nodes_from=[first, second],
                          log=log)

    # if you do not pass the log object, Chain will create default log.log file placed in core
    chain = Chain(final, log=log)

    return chain


if __name__ == '__main__':
    current_path = os.path.dirname(__name__)

    train_data, test_data = get_case_train_test_data()

    # Use default_log if you do not have json config file for log
    log = default_log('chain_log', log_file=os.path.join(current_path, 'example_log.log'))

    log.info('start creating chain')
    chain = get_simple_chain(log=log)

    log.info('start fitting chain')
    chain.fit(train_data, use_cache=False, verbose=True)



from copy import copy
from typing import List

import numpy as np
import pandas as pd

from fedot.core.models.data import (
    InputData,
)
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import extract_task_param


def ts_to_lagged_table(input_data: InputData) -> InputData:
    window_len, prediction_len = extract_task_param(input_data.task)

    transformed_data = copy(input_data)

    df = pd.DataFrame()
    df_target = None
    transformed_data.features = None

    if input_data.target is not None:
        target = list(input_data.target)
        target.append(np.nan)
        df = pd.DataFrame(data={'ts_target': target})
        df_target = pd.DataFrame(data={'ts_target': target})
        for lag in range(1, window_len + 1):
            df[f'lag_{lag}'] = df.ts_target.shift(lag)
        for lag in range(0, prediction_len):
            df_target[f'new_target_{lag}'] = df.ts_target.shift(-lag)

    if input_data.features is not None and (input_data.features.shape != input_data.target.shape or
                                            not np.allclose(input_data.features,
                                                            input_data.target,
                                                            equal_nan=True)):
        if len(input_data.features.shape) == 1:
            input_data.features.shape = input_data.features.shape + (1,)

        # iterate feature array by columns (inidvidual features)
        for feature_id, feature in enumerate(input_data.features.T):
            feature = list(feature)
            feature.append(np.nan)
            df_exog = pd.DataFrame(data={f'exog_feature': feature})
            for lag in range(1, window_len + 1):
                df[f'lag_f{feature_id}_{lag}'] = df_exog.exog_feature.shift(lag)

    df = df.drop('ts_target', axis=1)
    df_target = df_target.drop('ts_target', axis=1)

    transformed_data.features = np.squeeze(df)
    transformed_data.target = df_target

    size_diff = transformed_data.features.shape[0] - len(transformed_data.idx)
    if size_diff:
        new_idx = list(transformed_data.idx)
        new_idx.extend(['new'] * size_diff)
        transformed_data.idx = np.asarray(new_idx)

    transformed_data.data_type = DataTypesEnum.ts_lagged_table
    return transformed_data


def direct(input_data: InputData) -> InputData:
    return copy(input_data)


# from datatype / to datatype : function
_transformation_functions_for_data_types = {
    (DataTypesEnum.ts, DataTypesEnum.ts_lagged_table):
        ts_to_lagged_table,
    (DataTypesEnum.forecasted_ts, DataTypesEnum.table): direct,
    (DataTypesEnum.table, DataTypesEnum.ts): direct
}


def transformation_function_for_data(input_data_type: DataTypesEnum,
                                     required_data_types: List[DataTypesEnum]):
    if input_data_type in required_data_types:
        return direct

    transformation = None
    for required_data_type in required_data_types:
        transformation = _transformation_functions_for_data_types.get(
            (input_data_type, required_data_type), None)
        if transformation:
            break

    if not transformation:
        raise ValueError(
            f'The {input_data_type} cannot be converted to {required_data_types}')
    return transformation



from abc import ABC
from collections import namedtuple
from copy import copy
from datetime import timedelta
from typing import Callable, List, Optional

from fedot.core.log import default_log
from fedot.core.models.data import InputData, OutputData
from fedot.core.models.model import Model
from fedot.core.models.preprocessing import preprocessing_func_for_data
from fedot.core.models.transformation import transformation_function_for_data

CachedState = namedtuple('CachedState', 'preprocessor model')


class Node(ABC):
    """
    Base class for Node definition in Chain structure

    :param nodes_from: parent nodes which information comes from
    :param model_type: str type of the model defined in model repository
    :param manual_preprocessing_func: optional function for data preprocessing. \
    If not defined one of the available preprocessing strategies is used. \
    See the `preprocessors <https://github.com/nccr-itmo/FEDOT/blob/master/core/models/preprocessing.py>`__
    :param log: Log object to record messages
    """

    def __init__(self, nodes_from: Optional[List['Node']], model_type: str,
                 manual_preprocessing_func: Optional[Callable] = None,
                 log=None):
        self.nodes_from = nodes_from
        self.cache = FittedModelCache(self)
        self.manual_preprocessing_func = manual_preprocessing_func

        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

        self.model = Model(model_type=model_type)



    @property
    def descriptive_id(self):
        return self._descriptive_id_recursive(visited_nodes=[])

    def _descriptive_id_recursive(self, visited_nodes):
        node_label = self.model.description
        if self.manual_preprocessing_func:
            node_label = f'{node_label}_custom_preprocessing={self.manual_preprocessing_func.__name__}'
        full_path = ''
        if self in visited_nodes:
            return 'ID_CYCLED'
        visited_nodes.append(self)
        if self.nodes_from:
            previous_items = []
            for parent_node in self.nodes_from:
                previous_items.append(f'{parent_node._descriptive_id_recursive(copy(visited_nodes))};')
            previous_items.sort()
            previous_items_str = ';'.join(previous_items)

            full_path += f'({previous_items_str})'
        full_path += f'/{node_label}'
        return full_path

    @property
    def model_tags(self) -> List[str]:
        return self.model.metadata.tags

    def output_from_prediction(self, input_data, prediction):
        return OutputData(idx=input_data.idx,
                          features=input_data.features,
                          predict=prediction, task=input_data.task,
                          data_type=self.model.output_datatype(input_data.data_type))

    def _transform(self, input_data: InputData):
        transformed_data = transformation_function_for_data(
            input_data_type=input_data.data_type,
            required_data_types=self.model.metadata.input_types)(input_data)
        return transformed_data

    def _preprocess(self, data: InputData):
        preprocessing_func = preprocessing_func_for_data(data, self)

        if not self.cache.actual_cached_state:
            # if fitted preprocessor not found in cache
            preprocessing_strategy = \
                preprocessing_func().fit(data.features)
        else:
            # if fitted preprocessor already exists
            preprocessing_strategy = self.cache.actual_cached_state.preprocessor

        data.features = preprocessing_strategy.apply(data.features)

        return data, preprocessing_strategy

    def fit(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Run training process in the node

        :param input_data: data used for model training
        :param verbose: flag used for status printing to console, default False
        """
        transformed = self._transform(input_data)
        preprocessed_data, preproc_strategy = self._preprocess(transformed)

        if not self.cache.actual_cached_state:
            if verbose:
                print('Cache is not actual')

            cached_model, model_predict = self.model.fit(data=preprocessed_data)
            self.cache.append(CachedState(preprocessor=copy(preproc_strategy),
                                          model=cached_model))
        else:
            if verbose:
                print('Model were obtained from cache')

            model_predict = self.model.predict(fitted_model=self.cache.actual_cached_state.model,
                                               data=preprocessed_data)

        return self.output_from_prediction(input_data, model_predict)

    def predict(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Run prediction process in the node

        :param input_data: data used for prediction
        :param verbose: flag used for status printing to console, default False
        """
        transformed = self._transform(input_data)
        preprocessed_data, _ = self._preprocess(transformed)

        if not self.cache:
            raise ValueError('Model must be fitted before predict')

        model_predict = self.model.predict(fitted_model=self.cache.actual_cached_state.model,
                                           data=preprocessed_data)

        return self.output_from_prediction(input_data, model_predict)

    def fine_tune(self, input_data: InputData,
                  max_lead_time: timedelta = timedelta(minutes=5), iterations: int = 30):
        """
        Run the process of hyperparameter optimization for the node

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        """

        transformed = self._transform(input_data)
        preprocessed_data, preproc_strategy = self._preprocess(transformed)

        fitted_model, _ = self.model.fine_tune(preprocessed_data,
                                               max_lead_time=max_lead_time,
                                               iterations=iterations)

        self.cache.append(CachedState(preprocessor=copy(preproc_strategy),
                                      model=fitted_model))

    def __str__(self):
        model = f'{self.model}'
        return model

    @property
    def ordered_subnodes_hierarchy(self) -> List['Node']:
        nodes = [self]
        if self.nodes_from:
            for parent in self.nodes_from:
                nodes += parent.ordered_subnodes_hierarchy
        return nodes

    @property
    def custom_params(self) -> dict:
        return self.model.params

    @custom_params.setter
    def custom_params(self, params):
        if params:
            self.model.params = params


class FittedModelCache:
    def __init__(self, related_node: Node):
        self._local_cached_models = {}
        self._related_node_ref = related_node

    def append(self, fitted_model):
        self._local_cached_models[self._related_node_ref.descriptive_id] = fitted_model

    def import_from_other_cache(self, other_cache: 'FittedModelCache'):
        for entry_key in other_cache._local_cached_models.keys():
            self._local_cached_models[entry_key] = other_cache._local_cached_models[entry_key]

    def clear(self):
        self._local_cached_models = {}

    @property
    def actual_cached_state(self):
        found_model = self._local_cached_models.get(self._related_node_ref.descriptive_id, None)
        return found_model


class SharedCache(FittedModelCache):
    def __init__(self, related_node: Node, global_cached_models: dict):
        super().__init__(related_node)
        self._global_cached_models = global_cached_models

    def append(self, fitted_model):
        super().append(fitted_model)
        if self._global_cached_models is not None:
            self._global_cached_models[self._related_node_ref.descriptive_id] = fitted_model

    @property
    def actual_cached_state(self):
        found_model = super().actual_cached_state

        if not found_model and self._global_cached_models:
            found_model = self._global_cached_models.get(self._related_node_ref.descriptive_id, None)
        return found_model


class PrimaryNode(Node):
    """
    The class defines the interface of Primary nodes where initial task data is located

    :param model_type: str type of the model defined in model repository
    :param manual_preprocessing_func: optional function for data preprocessing.
    :param kwargs: optional arguments (i.e. logger)
    """

    def __init__(self, model_type: str, manual_preprocessing_func: Optional[Callable] = None,
                 **kwargs):
        super().__init__(nodes_from=None, model_type=model_type,
                         manual_preprocessing_func=manual_preprocessing_func,
                         **kwargs)

    def fit(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Fit the model located in the primary node

        :param input_data: data used for model training
        :param verbose: flag used for status printing to console, default False
        """
        if verbose:
            self.log.info(f'Trying to fit primary node with model: {self.model}')

        return super().fit(input_data, verbose)

    def predict(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Predict using the model located in the primary node

        :param input_data: data used for prediction
        :param verbose: flag used for status printing to console, default False
        """
        if verbose:
            self.log.info(f'Predict in primary node by model: {self.model}')

        return super().predict(input_data, verbose)


class SecondaryNode(Node):
    """
    The class defines the interface of Secondary nodes modifying tha data flow in Chain

    :param model_type: str type of the model defined in model repository
    :param nodes_from: parent nodes where data comes from
    :param manual_preprocessing_func: optional function for data preprocessing.
    :param kwargs: optional arguments (i.e. logger)
    """

    def __init__(self, model_type: str, nodes_from: Optional[List['Node']] = None,
                 manual_preprocessing_func: Optional[Callable] = None,
                 **kwargs):
        nodes_from = [] if nodes_from is None else nodes_from
        super().__init__(nodes_from=nodes_from, model_type=model_type,
                         manual_preprocessing_func=manual_preprocessing_func,
                         **kwargs)

    def fit(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Fit the model located in the secondary node

        :param input_data: data used for model training
        :param verbose: flag used for status printing to console, default False
        """
        if verbose:
            self.log.info(f'Trying to fit secondary node with model: {self.model}')

        secondary_input = self._input_from_parents(input_data=input_data,
                                                   parent_operation='fit',
                                                   verbose=verbose)
        return super().fit(input_data=secondary_input)

    def predict(self, input_data: InputData, verbose=False) -> OutputData:
        """
        Predict using the model located in the secondary node

        :param input_data: data used for prediction
        :param verbose: flag used for status printing to console, default False
        """
        if verbose:
            self.log.info(f'Obtain prediction in secondary node with model: {self.model}')

        secondary_input = self._input_from_parents(input_data=input_data,
                                                   parent_operation='predict',
                                                   verbose=verbose)

        return super().predict(input_data=secondary_input)

    def fine_tune(self, input_data: InputData, recursive: bool = True,
                  max_lead_time: timedelta = timedelta(minutes=5), iterations: int = 30,
                  verbose: bool = False):
        """
        Run the process of hyperparameter optimization for the node

        :param recursive: flag to initiate the tuning in the parent nodes or not, default: True
        :param input_data: data used for tuning
        :param max_lead_time: max time available for tuning process
        :param iterations: max number of iterations
        :param verbose: flag used for status printing to console, default True
        """
        if verbose:
            self.log.info(f'Tune all parent nodes in secondary node with model: {self.model}')

        if recursive:
            secondary_input = self._input_from_parents(input_data=input_data,
                                                       parent_operation='fine_tune',
                                                       max_tune_time=max_lead_time, verbose=verbose)
        else:
            secondary_input = self._input_from_parents(input_data=input_data,
                                                       parent_operation='fit',
                                                       max_tune_time=max_lead_time, verbose=verbose)

        return super().fine_tune(input_data=secondary_input)

    def _nodes_from_with_fixed_order(self):
        if self.nodes_from is not None:
            return sorted(self.nodes_from, key=lambda node: node.descriptive_id)

    def _input_from_parents(self, input_data: InputData,
                            parent_operation: str,
                            max_tune_time: Optional[timedelta] = None,
                            verbose=False) -> InputData:
        if len(self.nodes_from) == 0:
            raise ValueError()

        if verbose:
            self.log.info(f'Fit all parent nodes in secondary node with model: {self.model}')

        parent_nodes = self._nodes_from_with_fixed_order()

        are_prev_nodes_affect_target = \
            ['affects_target' in parent_node.model_tags for parent_node in parent_nodes]
        if any(are_prev_nodes_affect_target):
            # is the previous model is the model that changes target
            parent_results, target = _combine_parents_that_affects_target(parent_nodes, input_data,
                                                                          parent_operation)
        else:
            parent_results, target = _combine_parents_simple(parent_nodes, input_data,
                                                             parent_operation, max_tune_time)

        secondary_input = InputData.from_predictions(outputs=parent_results,
                                                     target=target)

        return secondary_input


def _combine_parents_that_affects_target(parent_nodes: List[Node],
                                         input_data: InputData,
                                         parent_operation: str):
    if len(parent_nodes) > 1:
        raise NotImplementedError()

    if parent_operation == 'predict':
        parent_result = parent_nodes[0].predict(input_data=input_data)
    elif parent_operation == 'fit' or parent_operation == 'fine_tune':
        parent_result = parent_nodes[0].fit(input_data=input_data)
    else:
        raise NotImplementedError()

    target = parent_result.predict
    return [parent_result], target


def _combine_parents_simple(parent_nodes: List[Node],
                            input_data: InputData,
                            parent_operation: str,
                            max_tune_time: Optional[timedelta]):
    target = input_data.target
    parent_results = []
    for parent in parent_nodes:
        if parent_operation == 'predict':
            prediction = parent.predict(input_data=input_data)
            parent_results.append(prediction)
        elif parent_operation == 'fit':
            prediction = parent.fit(input_data=input_data)
            parent_results.append(prediction)
        elif parent_operation == 'fine_tune':
            parent.fine_tune(input_data=input_data, max_lead_time=max_tune_time)
            prediction = parent.predict(input_data=input_data)
            parent_results.append(prediction)
        else:
            raise NotImplementedError()

    return parent_results, target



import os
from copy import deepcopy
from glob import glob, iglob
from math import ceil, log2
from os import remove
from time import time
from typing import Optional

import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd
from PIL import Image
from imageio import get_writer, imread

from fedot.core.composer.chain import Chain, as_nx_graph
from fedot.core.utils import default_fedot_data_dir


class ComposerVisualiser:
    default_data_dir = default_fedot_data_dir()
    temp_path = os.path.join(default_data_dir, 'composing_history')
    if 'composing_history' not in os.listdir(default_data_dir):
        os.mkdir(temp_path)
    gif_prefix = 'for_gif_'

    @staticmethod
    def visualise(chain: Chain, save_path: Optional[str] = None):
        try:
            graph, node_labels = as_nx_graph(chain=chain)
            pos = node_positions(graph.to_undirected())
            plt.figure(figsize=(10, 16))
            nx.draw(graph, pos=pos,
                    with_labels=True, labels=node_labels,
                    font_size=12, font_family='calibri', font_weight='bold',
                    node_size=7000, width=2.0,
                    node_color=colors_by_node_labels(node_labels), cmap='Set3')
            if not save_path:
                plt.show()
            else:
                plt.savefig(save_path)
        except Exception as ex:
            print(f'Visualisation failed with {ex}')

    @staticmethod
    def _visualise_chains(chains, fitnesses):
        fitnesses = deepcopy(fitnesses)
        last_best_chain = chains[0]

        prev_fit = fitnesses[0]

        for ch_id, chain in enumerate(chains):
            graph, node_labels = as_nx_graph(chain=chain)
            pos = node_positions(graph.to_undirected())
            plt.rcParams['axes.titlesize'] = 20
            plt.rcParams['axes.labelsize'] = 20
            plt.rcParams['figure.figsize'] = [10, 10]
            plt.title('Current chain')
            nx.draw(graph, pos=pos,
                    with_labels=True, labels=node_labels,
                    font_size=12, font_family='calibri', font_weight='bold',
                    node_size=scaled_node_size(chain.length), width=2.0,
                    node_color=colors_by_node_labels(node_labels), cmap='Set3')
            path = f'{ComposerVisualiser.temp_path}ch_{ch_id}.png'
            plt.savefig(path, bbox_inches='tight')

            plt.cla()
            plt.clf()
            plt.close('all')

            path_best = f'{ComposerVisualiser.temp_path}best_ch_{ch_id}.png'

            if fitnesses[ch_id] > prev_fit:
                fitnesses[ch_id] = prev_fit
            else:
                last_best_chain = chain
            prev_fit = fitnesses[ch_id]

            best_graph, best_node_labels = as_nx_graph(chain=last_best_chain)
            pos = node_positions(best_graph.to_undirected())
            plt.rcParams['axes.titlesize'] = 20
            plt.rcParams['axes.labelsize'] = 20
            plt.rcParams['figure.figsize'] = [10, 10]
            plt.title(f'Best chain after {round(ch_id)} evals')
            nx.draw(best_graph, pos=pos,
                    with_labels=True, labels=best_node_labels,
                    font_size=12, font_family='calibri', font_weight='bold',
                    node_size=scaled_node_size(chain.length), width=2.0,
                    node_color=colors_by_node_labels(best_node_labels), cmap='Set3')

            plt.savefig(path_best, bbox_inches='tight')

            plt.cla()
            plt.clf()
            plt.close('all')

    @staticmethod
    def _visualise_convergence(fitness_history):
        fitness_history = deepcopy(fitness_history)
        prev_fit = fitness_history[0]
        for fit_id, fit in enumerate(fitness_history):
            if fit > prev_fit:
                fitness_history[fit_id] = prev_fit
            prev_fit = fitness_history[fit_id]
        ts_set = list(range(len(fitness_history)))
        df = pd.DataFrame(
            {'ts': ts_set, 'fitness': [-f for f in fitness_history]})

        ind = 0
        for ts in ts_set:
            plt.rcParams['axes.titlesize'] = 20
            plt.rcParams['axes.labelsize'] = 20
            plt.rcParams['figure.figsize'] = [10, 10]

            ind = ind + 1
            plt.plot(df['ts'], df['fitness'], label='Composer')
            plt.xlabel('Evaluation', fontsize=18)
            plt.ylabel('Best ROC AUC', fontsize=18)

            plt.axvline(x=ts, color='black')
            plt.legend(loc='upper left')

            path = f'{ComposerVisualiser.temp_path}{ind}.png'
            plt.savefig(path, bbox_inches='tight')

            plt.cla()
            plt.clf()
            plt.close('all')

    @staticmethod
    def visualise_history(chains, fitnesses):
        print('START VISUALISATION')
        try:
            ComposerVisualiser._clean(with_gif=True)
            ComposerVisualiser._visualise_chains(chains, fitnesses)
            ComposerVisualiser._visualise_convergence(fitnesses)
            ComposerVisualiser._merge_images(len(chains))
            ComposerVisualiser._combine_gifs()
            ComposerVisualiser._clean()
        except Exception as ex:
            print(f'Visualisation failed with {ex}')

    @staticmethod
    def _merge_images(num_images):
        for img_idx in (range(1, num_images)):
            images = list(map(Image.open, [f'{ComposerVisualiser.temp_path}ch_{img_idx}.png',
                                           f'{ComposerVisualiser.temp_path}best_ch_{img_idx}.png',
                                           f'{ComposerVisualiser.temp_path}{img_idx}.png']))
            widths, heights = zip(*(i.size for i in images))

            total_width = sum(widths)
            max_height = max(heights)

            new_im = Image.new('RGB', (total_width, max_height))

            x_offset = 0
            for im in images:
                new_im.paste(im, (x_offset, 0))
                x_offset += im.size[0]

            new_im.save(f'{ComposerVisualiser.temp_path}{ComposerVisualiser.gif_prefix}{img_idx}.png')

    @staticmethod
    def _combine_gifs():
        files = [file_name for file_name in
                 iglob(f'{ComposerVisualiser.temp_path}{ComposerVisualiser.gif_prefix}*.png')]
        files_idx = [int(file_name[len(f'{ComposerVisualiser.temp_path}{ComposerVisualiser.gif_prefix}'):(
                len(file_name) - len('.png'))]) for
                     file_name in
                     iglob(f'{ComposerVisualiser.temp_path}{ComposerVisualiser.gif_prefix}*.png')]
        files = [file for _, file in sorted(zip(files_idx, files))]

        with get_writer(f'{ComposerVisualiser.temp_path}final_{str(time())}.gif', mode='I', duration=0.5) as writer:
            for filename in files:
                image = imread(filename)
                writer.append_data(image)

    @staticmethod
    def _clean(with_gif=False):
        try:
            files = glob(f'{ComposerVisualiser.temp_path}*.png')
            if with_gif:
                files += glob(f'{ComposerVisualiser.temp_path}*.gif')
            for file in files:
                remove(file)
        except Exception as ex:
            print(ex)


def colors_by_node_labels(node_labels: dict):
    colors = [color for color in range(len(node_labels.keys()))]
    return colors


def scaled_node_size(nodes_amount):
    size = int(7000.0 / max(ceil(log2(nodes_amount)), 1))
    return size


def node_positions(graph: nx.Graph):
    if not nx.is_tree(graph):
        raise TypeError('cannot use hierarchy_pos on a graph that is not a tree')
    return nx.drawing.nx_pydot.graphviz_layout(graph, prog='dot')



from typing import Dict

from sklearn import datasets


def classification_dataset(samples_amount: int, features_amount: int, classes_amount: int,
                           features_options: Dict, noise_fraction: float = 0.1,
                           full_shuffle: bool = True):
    """
    Generates a random dataset for n-class classification problem
    using scikit-learn API.
    :param samples_amount: Total amount of samples in the resulted dataset.
    :param features_amount: Total amount of features per sample.
    :param classes_amount: The amount of classes in the dataset.
    :param features_options: The dictionary containing features options in key-value
    format:
        - informative: the amount of informative features;
        - redundant: the amount of redundant features;
        - repeated: the amount of features that repeat the informative features;
        - clusters_per_class: the amount of clusters for each class;
    :param noise_fraction: the fraction of noisy labels in the dataset;
    :param full_shuffle: if true then all features and samples will be shuffled.
    :return: features and target as numpy-arrays.
    """
    features, target = datasets.make_classification(n_samples=samples_amount, n_features=features_amount,
                                                    n_informative=features_options['informative'],
                                                    n_redundant=features_options['redundant'],
                                                    n_repeated=features_options['repeated'],
                                                    n_classes=classes_amount,
                                                    n_clusters_per_class=features_options['clusters_per_class'],
                                                    flip_y=noise_fraction,
                                                    shuffle=full_shuffle)

    return features, target


def gauss_quantiles_dataset(samples_amount: int, features_amount: int,
                            classes_amount: int, full_shuffle=True, **kwargs):
    """
    Generates a random dataset for n-class classification problem
    based on multi-dimensional gaussian distribution quantiles
    using scikit-learn API.
    :param samples_amount: Total amount of samples in the resulted dataset.
    :param features_amount: Total amount of features per sample.
    :param classes_amount: The amount of classes in the dataset.
    :param full_shuffle: if true then all features and samples will be shuffled.
    :param kwargs: Optional params:
        - 'gauss_params': mean and covariance values of the distribution.
    :return: features and target as numpy-arrays.
    """
    if 'gauss_params' in kwargs:
        mean, cov = kwargs['gauss_params']
    else:
        mean, cov = None, 1.

    features, target = datasets.make_gaussian_quantiles(n_samples=samples_amount,
                                                        n_features=features_amount,
                                                        n_classes=classes_amount,
                                                        shuffle=full_shuffle,
                                                        mean=mean, cov=cov)
    return features, target



from typing import Dict

import numpy as np

from fedot.core.composer.chain import Chain
from fedot.core.models.data import InputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.utilities.synthetic.chain import separately_fit_chain
from fedot.utilities.synthetic.data import classification_dataset

DEFAULT_OPTIONS = {
    'informative': 10,
    'redundant': 0,
    'repeated': 0,
    'clusters_per_class': 1
}


def synthetic_benchmark_dataset(samples_amount: int, features_amount: int,
                                classes_amount: int = 2,
                                features_options: Dict = DEFAULT_OPTIONS,
                                fitted_chain: Chain = None) -> InputData:
    """
    Generates a binary classification benchmark dataset that was obtained using
    the (TODO: add. reference) proposed fitting schema.
    :param samples_amount: Total amount of samples in the resulted dataset.
    :param features_amount: Total amount of features per sample.
    :param classes_amount: The amount of classes in the dataset.
    :param features_options: features options in key-value suitable for classification_dataset.
    :param fitted_chain: Chain with separately fitted models.
    If None then 3-level balanced tree were fitted and taken as a default.
    :return: Benchmark dataset that is ready to be used by Chain.
    """
    if fitted_chain is None:
        fitted_chain = _default_chain(samples_amount=samples_amount,
                                      features_amount=features_amount,
                                      classes_amount=classes_amount)

    if classes_amount != 2:
        raise NotImplementedError('Only binary classification tasks are supported')

    features, target = classification_dataset(samples_amount=samples_amount,
                                              features_amount=features_amount,
                                              classes_amount=classes_amount,
                                              features_options=features_options)
    target = np.expand_dims(target, axis=1)

    task = Task(TaskTypesEnum.classification)
    samples_idxs = np.arange(0, samples_amount)

    train = InputData(idx=samples_idxs,
                      features=features, target=target, task=task,
                      data_type=DataTypesEnum.table)

    synth_target = fitted_chain.predict(input_data=train).predict
    synth_labels = _to_labels(synth_target)
    data_synth_train = InputData(idx=np.arange(0, samples_amount),
                                 features=features, target=synth_labels, task=task,
                                 data_type=DataTypesEnum.table)

    # TODO: fix preproc issues

    fitted_chain.fit_from_scratch(input_data=data_synth_train)

    features, target = classification_dataset(samples_amount=samples_amount,
                                              features_amount=features_amount,
                                              classes_amount=classes_amount,
                                              features_options=features_options)
    target = np.expand_dims(target, axis=1)
    test = InputData(idx=samples_idxs,
                     features=features, target=target,
                     data_type=DataTypesEnum.table, task=task)
    synth_target = fitted_chain.predict(input_data=test).predict
    synth_labels = _to_labels(synth_target)
    data_synth_final = InputData(idx=samples_idxs,
                                 features=features,
                                 data_type=DataTypesEnum.table,
                                 target=synth_labels, task=task)

    return data_synth_final


def _default_chain(samples_amount: int, features_amount: int,
                   classes_amount: int = 2):
    chain = separately_fit_chain(samples=samples_amount, features_amount=features_amount,
                                 classes=classes_amount)
    return chain


def _to_labels(predictions):
    thr = 0.5
    labels = [0 if val <= thr else 1 for val in predictions]
    labels = np.expand_dims(np.array(labels), axis=1)
    return labels



import os

import pytest

from fedot.core.log import Log, default_log
from fedot.core.models.data import train_test_data_setup, InputData
from fedot.core.models.model import Model
from test.test_chain_import_export import create_four_depth_chain
from fedot.core.log import LogManager


@pytest.fixture()
def get_config_file():
    test_file_path = str(os.path.dirname(__file__))
    file = os.path.join(test_file_path, 'data', 'logging.json')
    if os.path.exists(file):
        return file


def release_log(logger, log_file):
    logger.release_handlers()
    if os.path.exists(log_file):
        os.remove(log_file)


def test_default_logger_setup_correctly():
    expected_logger_info_level = 20
    test_default_log = default_log('default_test_logger')

    assert test_default_log.logger.getEffectiveLevel() == expected_logger_info_level


@pytest.mark.parametrize('data_fixture', ['get_config_file'])
def test_logger_from_config_file_setup_correctly(data_fixture, request):
    expected_logger_error_level = 40
    test_config_file = request.getfixturevalue(data_fixture)
    log = Log('test_logger', config_json_file=test_config_file)

    assert log.logger.getEffectiveLevel() == expected_logger_error_level


def test_logger_write_logs_correctly():
    test_file_path = str(os.path.dirname(__file__))
    test_log_file = os.path.join(test_file_path, 'test_log.log')
    test_log = default_log('test_log',
                           log_file=test_log_file)

    # Model data preparation
    file = os.path.join('data', 'advanced_classification.csv')
    data = InputData.from_csv(os.path.join(test_file_path, file))
    train_data, test_data = train_test_data_setup(data=data)

    try:
        knn = Model(model_type='knnreg', log=test_log)
        model, _ = knn.fit(data=train_data)
    except Exception:
        print('Captured error')

    if os.path.exists(test_log_file):
        with open(test_log_file, 'r') as file:
            content = file.readlines()

    release_log(logger=test_log, log_file=test_log_file)
    assert 'Can not find evaluation strategy' in content[0]


def test_logger_manager_keeps_loggers_correctly():
    LogManager().clear_cache()

    chain = create_four_depth_chain()
    expected_number_of_loggers = 4

    file = os.path.join('data', 'advanced_classification.csv')
    test_file_path = str(os.path.dirname(__file__))
    data = InputData.from_csv(os.path.join(test_file_path, file))
    train_data, _ = train_test_data_setup(data=data)

    chain.fit(train_data)

    actual_number_of_loggers = LogManager().debug['loggers_number']

    assert actual_number_of_loggers == expected_number_of_loggers



from dataclasses import dataclass
from typing import (
    Callable,
    Optional,
)

from fedot.core.composer.chain import Chain
from fedot.core.composer.composer import ComposerRequirements
from fedot.core.composer.gp_composer.gp_composer import GPComposer, GPComposerBuilder
from fedot.core.composer.optimisers.gp_optimiser import GPChainOptimiserParameters
from fedot.core.composer.optimisers.mutation import MutationTypesEnum
from fedot.core.models.data import InputData
from fedot.core.repository.tasks import Task


@dataclass
class GPComposerRequirements(ComposerRequirements):
    pop_size: Optional[int] = 50
    num_of_generations: Optional[int] = 50
    crossover_prob: Optional[float] = None
    mutation_prob: Optional[float] = None


class FixedStructureComposer(GPComposer):
    def __init__(self, optimiser=None, metrics: Optional[Callable] = None,
                 composer_requirements: GPComposerRequirements = None,
                 initial_chain: Optional[Chain] = None):
        super().__init__(optimiser=optimiser, metrics=metrics, composer_requirements=composer_requirements,
                         initial_chain=initial_chain)

    def compose_chain(self, data: InputData, is_visualise: bool = False) -> Chain:
        return super().compose_chain(data, is_visualise)


class FixedStructureComposerBuilder(GPComposerBuilder):
    def __init__(self, task: Task):
        super().__init__(task=task)
        self._composer = FixedStructureComposer()
        fixed_structure_optimiser_parameters = GPChainOptimiserParameters(mutation_types=[MutationTypesEnum.simple])
        self.optimiser_parameters = fixed_structure_optimiser_parameters

    def set_default_composer_params(self):
        super().set_default_composer_params()
        self._composer.composer_requirements.crossover_prob = 0.0



import json
import os
import warnings
from dataclasses import dataclass
from typing import Any, List, Optional

from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.json_evaluation import eval_field_str, eval_strategy_str, read_field
from fedot.core.repository.tasks import TaskTypesEnum


@dataclass
class ModelMetaInfo:
    id: str
    input_types: List[DataTypesEnum]
    output_types: List[DataTypesEnum]
    task_type: List[TaskTypesEnum]
    supported_strategies: Any
    allowed_positions: List[str]
    tags: Optional[List[str]] = None

    def current_strategy(self, task: TaskTypesEnum):
        if isinstance(self.supported_strategies, dict):
            return self.supported_strategies.get(task, None)
        return self.supported_strategies


class ModelTypesRepository:
    _repo = None

    def __init__(self, repo_path=None):
        if repo_path:
            ModelTypesRepository._repo = self._initialise_repo(repo_path)
        if not repo_path and not ModelTypesRepository._repo:
            self._set_repo_to_default_state()

        self._tags_excluded_by_default = ['non-default', 'expensive']

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._set_repo_to_default_state()

    def _set_repo_to_default_state(self):
        repo_folder_path = str(os.path.dirname(__file__))
        file = 'data/model_repository.json'
        repo_path = os.path.join(repo_folder_path, file)
        ModelTypesRepository._repo = self._initialise_repo(repo_path)

    def _initialise_repo(self, repo_path: str) -> List[ModelMetaInfo]:
        with open(repo_path) as repository_json_file:
            repository_json = json.load(repository_json_file)

        metadata_json = repository_json['metadata']
        models_json = repository_json['models']

        models_list = []

        for current_model_key in models_json:
            model_properties = [model_properties for model_key, model_properties in list(models_json.items())
                                if model_key == current_model_key][0]
            model_metadata = metadata_json[model_properties['meta']]

            task_types = eval_field_str(model_metadata['tasks'])
            input_type = eval_field_str(model_metadata['input_type'])
            output_type = eval_field_str(model_metadata['output_type'])

            strategies_json = model_metadata['strategies']
            if isinstance(strategies_json, list):
                supported_strategies = eval_strategy_str(strategies_json)
            else:
                supported_strategies = {}
                for strategy_dict_key in strategies_json.keys():
                    supported_strategies[eval_field_str(strategy_dict_key)] = eval_strategy_str(
                        strategies_json[strategy_dict_key])

            accepted_node_types = read_field(model_metadata, 'accepted_node_types', ['any'])
            forbidden_node_types = read_field(model_metadata, 'forbidden_node_types', [])
            meta_tags = read_field(model_metadata, 'tags', [])

            model_tags = read_field(model_properties, 'tags', [])

            allowed_positions = ['primary', 'secondary', 'root']

            if accepted_node_types and accepted_node_types != 'all':
                allowed_positions = accepted_node_types
            if forbidden_node_types:
                allowed_positions = [pos for pos in allowed_positions if pos not in forbidden_node_types]

            tags = list(set(meta_tags + model_tags))

            model = ModelMetaInfo(id=current_model_key,
                                  input_types=input_type, output_types=output_type, task_type=task_types,
                                  supported_strategies=supported_strategies,
                                  allowed_positions=allowed_positions,
                                  tags=tags)
            models_list.append(model)

        return models_list

    @property
    def models(self):
        return ModelTypesRepository._repo

    def model_info_by_id(self, id: str) -> Optional[ModelMetaInfo]:
        models_with_id = [m for m in ModelTypesRepository._repo if m.id == id]
        if len(models_with_id) > 1:
            raise ValueError('Several models with same id in repository')
        if len(models_with_id) == 0:
            warnings.warn('Model {id} not found in the repository')
            return None
        return models_with_id[0]

    def models_with_tag(self, tags: List[str], is_full_match: bool = False):
        models_info = [m for m in ModelTypesRepository._repo if
                       _is_tags_contains_in_model(tags, m.tags, is_full_match)]
        return [m.id for m in models_info], models_info

    def suitable_model(self, task_type: TaskTypesEnum,
                       tags: List[str] = None, is_full_match: bool = False,
                       forbidden_tags: List[str] = None):

        if not forbidden_tags:
            forbidden_tags = []

        for excluded_default_tag in self._tags_excluded_by_default:
            if not tags or excluded_default_tag not in tags:
                forbidden_tags.append(excluded_default_tag)

        models_info = [m for m in ModelTypesRepository._repo if task_type in m.task_type and
                       (not tags or _is_tags_contains_in_model(tags, m.tags, is_full_match)) and
                       (not forbidden_tags or not _is_tags_contains_in_model(forbidden_tags, m.tags, False))]
        return [m.id for m in models_info], models_info


def _is_tags_contains_in_model(candidate_tags: List[str], model_tags: List[str], is_full_match: bool):
    matches = ([(tag in model_tags) for tag in candidate_tags])
    if is_full_match:
        return all(matches)
    else:
        return any(matches)



import math
from copy import deepcopy
from functools import partial
from typing import (Any, Callable, List, Optional, Tuple)

import numpy as np

from fedot.core.composer.constraint import constraint_function
from fedot.core.composer.optimisers.crossover import CrossoverTypesEnum, crossover
from fedot.core.composer.optimisers.gp_operators import random_chain, num_of_parents_in_crossover
from fedot.core.composer.optimisers.inheritance import GeneticSchemeTypesEnum, inheritance
from fedot.core.composer.optimisers.mutation import MutationTypesEnum, mutation
from fedot.core.composer.optimisers.regularization import RegularizationTypesEnum, regularized_population
from fedot.core.composer.optimisers.selection import SelectionTypesEnum, selection
from fedot.core.composer.timer import CompositionTimer
from fedot.core.log import default_log, Log


class GPChainOptimiserParameters:
    """
        This class is for defining the parameters of optimiser

        :param selection_types: List of selection operators types
        :param crossover_types: List of crossover operators types
        :param mutation_types: List of mutation operators types
        :param regularization_type: type of regularization operator
        :param genetic_scheme_type: type of genetic evolutionary scheme
        :param with_auto_depth_configuration: flag to enable option of automated tree depth configuration during
        evolution. Default False.
        :param depth_increase_step: the step of depth increase in automated depth configuration
        :param start_depth: start value of tree depth. Using when with_auto_depth_configuration is True
    """

    def __init__(self, selection_types: List[SelectionTypesEnum] = None,
                 crossover_types: List[CrossoverTypesEnum] = None,
                 mutation_types: List[MutationTypesEnum] = None,
                 regularization_type: RegularizationTypesEnum = RegularizationTypesEnum.none,
                 genetic_scheme_type: GeneticSchemeTypesEnum = GeneticSchemeTypesEnum.generational,
                 with_auto_depth_configuration: bool = False, depth_increase_step: int = 3,
                 start_depth: int = 3):

        self.selection_types = selection_types
        self.crossover_types = crossover_types
        self.mutation_types = mutation_types
        self.regularization_type = regularization_type
        self.genetic_scheme_type = genetic_scheme_type
        self.with_auto_depth_configuration = with_auto_depth_configuration
        self.depth_increase_step = depth_increase_step
        self.start_depth = start_depth
        self.set_default_params()

    def set_default_params(self):
        if not self.selection_types:
            self.selection_types = [SelectionTypesEnum.tournament]
        if not self.crossover_types:
            self.crossover_types = [CrossoverTypesEnum.subtree]
        if not self.mutation_types:
            self.mutation_types = [MutationTypesEnum.simple]


class GPChainOptimiser:
    """
    Base class of evolutionary chain optimiser

    :param initial_chain: chain which was initialized outside the optimiser
    :param requirements: composer requirements
    :param chain_generation_params: parameters for new chain generation
    :param parameters: parameters of chain optimiser
    :param log: optional parameter for log oject
    """

    def __init__(self, initial_chain, requirements, chain_generation_params,
                 parameters: Optional[GPChainOptimiserParameters] = None, log: Log = None):
        self.chain_generation_params = chain_generation_params
        self.primary_node_func = self.chain_generation_params.primary_node_func
        self.secondary_node_func = self.chain_generation_params.secondary_node_func
        self.chain_class = self.chain_generation_params.chain_class
        self.requirements = requirements
        self.history = []
        self.parameters = GPChainOptimiserParameters() if parameters is None else parameters
        self.max_depth = self.parameters.start_depth if self.parameters.with_auto_depth_configuration else \
            self.requirements.max_depth

        self.generation_num = 0
        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

        self.chain_generation_function = partial(random_chain, chain_generation_params=self.chain_generation_params,
                                                 requirements=self.requirements, max_depth=self.max_depth)

        necessary_attrs = ['add_node', 'root_node', 'replace_node_with_parents', 'update_node', 'node_childs']
        if not all([hasattr(self.chain_class, attr) for attr in necessary_attrs]):
            ex = f'Object chain_class has no required attributes for gp_optimizer'
            self.log.error(ex)
            raise AttributeError(ex)

        if not self.requirements.pop_size:
            self.requirements.pop_size = 10

        if initial_chain and type(initial_chain) != list:
            self.population = [deepcopy(initial_chain) for _ in range(self.requirements.pop_size)]
        else:
            self.population = initial_chain

    def optimise(self, objective_function, offspring_rate: float = 0.5):
        if self.population is None:
            self.population = self._make_population(self.requirements.pop_size)

        num_of_new_individuals = self.offspring_size(offspring_rate)

        with CompositionTimer() as t:

            self.history = []

            if self.requirements.add_single_model_chains:
                best_single_model, self.requirements.primary = \
                    self._best_single_models(objective_function)

            for ind in self.population:
                ind.fitness = objective_function(ind)

            self._add_to_history(self.population)

            self.log.info(f'Best metric is {self.best_individual.fitness}')

            for self.generation_num in range(self.requirements.num_of_generations - 1):
                self.log.info(f'Generation num: {self.generation_num}')
                self.num_of_gens_without_improvements = self.update_stagnation_counter()
                self.log.info(
                    f'max_depth: {self.max_depth}, no improvements: {self.num_of_gens_without_improvements}')

                if self.parameters.with_auto_depth_configuration and self.generation_num != 0:
                    self.max_depth_recount()

                individuals_to_select = regularized_population(reg_type=self.parameters.regularization_type,
                                                               population=self.population,
                                                               objective_function=objective_function,
                                                               chain_class=self.chain_class)

                num_of_parents = num_of_parents_in_crossover(num_of_new_individuals)

                selected_individuals = selection(types=self.parameters.selection_types,
                                                 population=individuals_to_select,
                                                 pop_size=num_of_parents)

                new_population = []

                for parent_num in range(0, len(selected_individuals), 2):
                    new_population += self.reproduce(selected_individuals[parent_num],
                                                     selected_individuals[parent_num + 1])

                    new_population[parent_num].fitness = objective_function(new_population[parent_num])
                    new_population[parent_num + 1].fitness = objective_function(new_population[parent_num + 1])

                self.prev_best = deepcopy(self.best_individual)

                self.population = inheritance(self.parameters.genetic_scheme_type, self.parameters.selection_types,
                                              self.population,
                                              new_population, self.num_of_inds_in_next_pop)

                if self.with_elitism:
                    self.population.append(self.prev_best)

                self._add_to_history(self.population)
                self.log.info(f'spent time: {round(t.minutes_from_start, 1)} min')
                self.log.info(f'Best metric is {self.best_individual.fitness}')

                if t.is_time_limit_reached(self.requirements.max_lead_time, self.generation_num):
                    break

            best = self.best_individual

            if self.requirements.add_single_model_chains and \
                    (best_single_model.fitness <= best.fitness):
                best = best_single_model
        return best, self.history

    @property
    def best_individual(self) -> Any:
        return self.get_best_individual(self.population)

    @property
    def with_elitism(self) -> bool:
        return self.requirements.pop_size > 1

    @property
    def num_of_inds_in_next_pop(self):
        return self.requirements.pop_size - 1 if self.with_elitism else self.requirements.pop_size

    def update_stagnation_counter(self) -> int:
        value = 0
        if self.generation_num != 0:
            if self.is_equal_fitness(self.prev_best.fitness, self.best_individual.fitness):
                value = self.num_of_gens_without_improvements + 1

        return value

    def max_depth_recount(self):
        if self.num_of_gens_without_improvements == self.parameters.depth_increase_step and \
                self.max_depth + 1 <= self.requirements.max_depth:
            self.max_depth += 1

    def get_best_individual(self, individuals: List[Any], equivalents_from_current_pop=True) -> Any:
        best_ind = min(individuals, key=lambda ind: ind.fitness)
        if equivalents_from_current_pop:
            equivalents = self.simpler_equivalents_of_best_ind(best_ind)
        else:
            equivalents = self.simpler_equivalents_of_best_ind(best_ind, individuals)

        if equivalents:
            best_candidate_id = min(equivalents, key=equivalents.get)
            best_ind = individuals[best_candidate_id]
        return best_ind

    def simpler_equivalents_of_best_ind(self, best_ind: Any, inds: List[Any] = None) -> dict:
        individuals = self.population if inds is None else inds

        sort_inds = np.argsort([ind.fitness for ind in individuals])[1:]
        simpler_equivalents = {}
        for i in sort_inds:
            is_fitness_equals_to_best = self.is_equal_fitness(best_ind.fitness, individuals[i].fitness)
            has_less_num_of_models_than_best = len(individuals[i].nodes) < len(best_ind.nodes)
            if is_fitness_equals_to_best and has_less_num_of_models_than_best:
                simpler_equivalents[i] = len(individuals[i].nodes)
        return simpler_equivalents

    def reproduce(self, selected_individual_first, selected_individual_second=None) -> Tuple[Any]:
        if selected_individual_second:
            new_inds = crossover(self.parameters.crossover_types,
                                 selected_individual_first,
                                 selected_individual_second,
                                 crossover_prob=self.requirements.crossover_prob,
                                 max_depth=self.requirements.max_depth)
        else:
            new_inds = [selected_individual_first]

        new_inds = tuple([mutation(types=self.parameters.mutation_types,
                                   chain_generation_params=self.chain_generation_params,
                                   chain=new_ind, requirements=self.requirements,
                                   max_depth=self.max_depth) for new_ind in new_inds])

        return new_inds

    def _make_population(self, pop_size: int) -> List[Any]:
        model_chains = []
        while len(model_chains) < pop_size:
            chain = self.chain_generation_function()
            if constraint_function(chain):
                model_chains.append(chain)
        return model_chains

    def _add_to_history(self, individuals: List[Any]):
        self.history.append(individuals)

    def _best_single_models(self, objective_function: Callable, num_best: int = 7):
        single_models_inds = []
        for model in self.requirements.primary:
            single_models_ind = self.chain_class([self.primary_node_func(model)])
            single_models_ind.fitness = objective_function(single_models_ind)
            single_models_inds.append(single_models_ind)
        best_inds = sorted(single_models_inds, key=lambda ind: ind.fitness)
        return best_inds[0], [i.nodes[0].model.model_type for i in best_inds][:num_best]

    def offspring_size(self, offspring_rate: float = None):
        default_offspring_rate = 0.5 if not offspring_rate else offspring_rate
        if self.parameters.genetic_scheme_type == GeneticSchemeTypesEnum.steady_state:
            num_of_new_individuals = math.ceil(self.requirements.pop_size * default_offspring_rate)
        else:
            num_of_new_individuals = self.requirements.pop_size - 1
        return num_of_new_individuals

    def is_equal_fitness(self, first_fitness, second_fitness, atol=1e-10, rtol=1e-10):
        return np.isclose(first_fitness, second_fitness, atol=atol, rtol=rtol)



import os
from os.path import dirname, join
from pathlib import Path

import setuptools

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
NAME = "fedot"
VERSION = "0.1.3"
AUTHOR = "NSS Lab"
SHORT_DESCRIPTION = "Evolutionary structural learning framework FEDOT"
README = (Path(os.path.join(HERE, "README.rst"))).read_text()
URL = "https://github.com/nccr-itmo/FEDOT"
REQUIRES_PYTHON = '>=3.6'
LICENSE = "BSD 3-Clause"


def read(*names, **kwargs):
    with open(
            join(dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


def extract_requirements(file_name):
    return [r for r in read(file_name).split('\n') if r and not r.startswith('#')]


def get_requirements():
    requirements = extract_requirements('requirements.txt')
    return requirements


setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email="itmo.nss.team@gmail.com",
    description=SHORT_DESCRIPTION,
    long_description=README,
    long_description_content_type="text/x-rst",
    url=URL,
    python_requires=REQUIRES_PYTHON,
    license=LICENSE,
    packages=setuptools.find_packages(exclude=['test*']),
    include_package_data=True,
    install_requires=get_requirements(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from fedot.core.composer.iterator import fibonacci_sequence, SequenceIterator


def test_iterator_without_constraints():
    sequence_generation_function = fibonacci_sequence
    iterator = SequenceIterator(sequence_func=sequence_generation_function)
    num_of_values = 5
    sequence_values = [iterator.next() for _ in range(num_of_values)]
    assert sequence_values[0] == 0 and sequence_values[4] == 3


def test_iterator_from_certain_index():
    sequence_generation_function = fibonacci_sequence
    iterator = SequenceIterator(sequence_func=sequence_generation_function, start_value=3)
    num_of_values = 5
    sequence_values = [iterator.next() for _ in range(num_of_values)]
    assert sequence_values[0] == 3 and sequence_values[4] == 21


def test_iterator_with_max_min_constraints():
    sequence_generation_function = fibonacci_sequence
    iterator = SequenceIterator(sequence_func=sequence_generation_function, min_sequence_value=10,
                                max_sequence_value=25)
    assert not iterator.has_prev()
    first_value = iterator.next()
    second_value = iterator.next()
    assert first_value == 13
    assert second_value == 21
    assert not iterator.has_next()
    assert iterator.has_prev()



import os
from enum import Enum
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split


def project_root() -> Path:
    """Returns project root folder."""
    return Path(__file__).parent.parent.parent


def default_fedot_data_dir() -> str:
    """ Returns the folder where all the output data
    is recorded to. Default: home/Fedot
    """
    default_data_path = os.path.join(str(Path.home()), 'Fedot')
    if 'Fedot' not in os.listdir(str(Path.home())):
        os.mkdir(default_data_path)
    return default_data_path


def labels_to_dummy_probs(prediction: np.array):
    """Returns converted predictions
    using one-hot probability encoding"""
    df = pd.Series(prediction)
    pred_probas = pd.get_dummies(df).values
    return pred_probas


def probs_to_labels(prediction: np.array):
    list_with_labels = []
    for list_with_probs in prediction:
        list_with_labels.append(list_with_probs.argmax() + 1.0)

    return list_with_labels


def ensure_features_2d(features: np.array):
    if len(features.shape) >= 3:
        num_of_samples = features.shape[1]
        features_2d = features.reshape(num_of_samples, -1)
        return features_2d
    else:
        return features


def split_data(df: pd.DataFrame, t_size: float = 0.2):
    train, test = train_test_split(df.iloc[:, :], test_size=t_size, random_state=42)
    return train, test


def save_file_to_csv(df: pd.DataFrame, path_to_save: str):
    return df.to_csv(path_to_save, sep=',')


def get_split_data_paths(directory_names: list):
    train_file_path = os.path.join(directory_names[0], directory_names[1], directory_names[2], 'train.csv')
    full_train_file_path = os.path.join(str(project_root()), train_file_path)
    test_file_path = os.path.join(directory_names[0], directory_names[1], directory_names[2], 'test.csv')
    full_test_file_path = os.path.join(str(project_root()), test_file_path)
    return full_train_file_path, full_test_file_path


def ensure_directory_exists(dir_names: list):
    main_dir = os.path.join(str(project_root()), dir_names[0], dir_names[1])
    dataset_dir = os.path.join(str(project_root()), dir_names[0], dir_names[1], dir_names[2])
    if not os.path.exists(main_dir):
        os.mkdir(main_dir)
    if not os.path.exists(dataset_dir):
        os.mkdir(dataset_dir)


class ComparableEnum(Enum):
    """
    The Enum implementation that allows to avoid the multi-module enum comparison problem
    (https://stackoverflow.com/questions/26589805/python-enums-across-modules)
    """

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))



from typing import Any, List, Optional

from dataclasses import dataclass

from fedot.core.utils import ComparableEnum as Enum


@dataclass
class TaskParams:
    pass


@dataclass
class TsForecastingParams(TaskParams):
    forecast_length: int
    max_window_size: int
    return_all_steps: bool = False
    make_future_prediction: bool = False


class TaskTypesEnum(Enum):
    classification = 'classification',
    regression = 'regression',
    ts_forecasting = 'ts_forecasting'
    clustering = 'clustering'  # not applicable as main task yet


@dataclass
class Task:
    task_type: TaskTypesEnum
    task_params: Optional[TaskParams] = None


# local tasks that can be solved as a part of global tasks
def compatible_task_types(main_task_type: TaskTypesEnum) -> List[TaskTypesEnum]:
    _compatible_task_types = {
        TaskTypesEnum.ts_forecasting: [TaskTypesEnum.regression],
        TaskTypesEnum.classification: [TaskTypesEnum.clustering],
        TaskTypesEnum.regression: [TaskTypesEnum.clustering]
    }
    if main_task_type not in _compatible_task_types:
        return []
    return _compatible_task_types[main_task_type]


def extract_task_param(task: Task) -> Any:
    try:
        task_params = task.task_params
        if isinstance(task_params, TsForecastingParams):
            window_len = task_params.max_window_size
            prediction_len = task_params.forecast_length
            return window_len, prediction_len
        else:
            raise ValueError('Incorrect parameters type for data')
    except AttributeError as ex:
        raise AttributeError(f'Params are required for the {task} task: {ex}')



import numpy as np
from scipy import interpolate

from fedot.core.models.data import InputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams


class SimpleGapFiller:
    """
    Base class used for filling in the gaps in time series with simple methods.
    Methods from the SimpleGapFiller class can be used for comparison with more
    complex models in class ModelGapFiller

    :param gap_value: value, which identify gap elements in array
    """

    def __init__(self, gap_value: float = -100.0):
        self.gap_value = gap_value

    def linear_interpolation(self, input_data):
        """
        Method allows to restore missing values in an array
        using linear interpolation

        :param input_data: array with gaps
        :return: array without gaps
        """

        output_data = np.array(input_data)

        # The indices of the known elements
        non_nan = np.ravel(np.argwhere(output_data != self.gap_value))
        # All known elements in the array
        masked_array = output_data[non_nan]
        f_interploate = interpolate.interp1d(non_nan, masked_array)
        x = np.arange(0, len(output_data))
        output_data = f_interploate(x)
        return output_data

    def local_poly_approximation(self, input_data, degree: int = 2,
                                 n_neighbors: int = 5):
        """
        Method allows to restore missing values in an array
        using Savitzky-Golay filter

        :param input_data: array with gaps
        :param degree: degree of a polynomial function
        :param n_neighbors: number of neighboring known elements of the time
        series that the approximation is based on
        :return: array without gaps
        """

        output_data = np.array(input_data)

        i_gaps = np.ravel(np.argwhere(output_data == self.gap_value))

        # Iterately fill in the gaps in the time series
        for gap_index in i_gaps:
            # Indexes of known elements (updated at each iteration)
            i_known = np.argwhere(output_data != self.gap_value)
            i_known = np.ravel(i_known)

            # Based on the indexes we calculate how far from the gap
            # the known values are located
            id_distances = np.abs(i_known - gap_index)

            # Now we know the indices of the smallest values in the array,
            # so sort indexes
            sorted_idx = np.argsort(id_distances)
            nearest_values = []
            nearest_indices = []
            for i in sorted_idx[:n_neighbors]:
                time_index = i_known[i]
                nearest_values.append(output_data[time_index])
                nearest_indices.append(time_index)
            nearest_values = np.array(nearest_values)
            nearest_indices = np.array(nearest_indices)

            local_coefs = np.polyfit(nearest_indices, nearest_values, degree)
            est_value = np.polyval(local_coefs, gap_index)
            output_data[gap_index] = est_value

        return output_data

    def batch_poly_approximation(self, input_data, degree: int = 3,
                                 n_neighbors: int = 10):
        """
        Method allows to restore missing values in an array using
        batch polynomial approximations.
        Approximation is applied not for individual omissions, but for
        intervals of omitted values

        :param input_data: array with gaps
        :param degree: degree of a polynomial function
        :param n_neighbors: the number of neighboring known elements of
        time series that the approximation is based on
        :return: array without gaps
        """

        output_data = np.array(input_data)

        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iterately fill in the gaps in the time series
        for gap in new_gap_list:
            # Find the center point of the gap
            center_index = int((gap[0] + gap[-1]) / 2)

            # Indexes of known elements (updated at each iteration)
            i_known = np.argwhere(output_data != self.gap_value)
            i_known = np.ravel(i_known)

            # Based on the indexes we calculate how far from the gap
            # the known values are located
            id_distances = np.abs(i_known - center_index)

            # Now we know the indices of the smallest values in the array,
            # so sort indexes
            sorted_idx = np.argsort(id_distances)

            # Nearest known values to the gap
            nearest_values = []
            # And their indexes
            nearest_indices = []
            for i in sorted_idx[:n_neighbors]:
                # Getting the index value for the series - output_data
                time_index = i_known[i]
                # Using this index, we get the value of each of the "neighbors"
                nearest_values.append(output_data[time_index])
                nearest_indices.append(time_index)
            nearest_values = np.array(nearest_values)
            nearest_indices = np.array(nearest_indices)

            # Local approximation by an n-th degree polynomial
            local_coefs = np.polyfit(nearest_indices, nearest_values, degree)

            # Estimate our interval according to the selected coefficients
            est_value = np.polyval(local_coefs, gap)
            output_data[gap] = est_value

        return output_data

    def _parse_gap_ids(self, gap_list: list) -> list:
        """
        Method allows parsing source array with gaps indexes

        :param gap_list: array with indexes of gaps in array
        :return: a list with separated gaps in continuous intervals
        """

        new_gap_list = []
        local_gaps = []
        for index, gap in enumerate(gap_list):
            if index == 0:
                local_gaps.append(gap)
            else:
                prev_gap = gap_list[index - 1]
                if gap - prev_gap > 1:
                    # There is a "gap" between gaps
                    new_gap_list.append(local_gaps)

                    local_gaps = []
                    local_gaps.append(gap)
                else:
                    local_gaps.append(gap)
        new_gap_list.append(local_gaps)

        return new_gap_list


class ModelGapFiller(SimpleGapFiller):
    """
    Class used for filling in the gaps in time series

    :param gap_value: value, which mask gap elements in array
    :param chain: TsForecastingChain object for filling in the gaps
    :param max_window_size: window length
    """

    def __init__(self, gap_value, chain, max_window_size: int = 50):
        super().__init__(gap_value)
        self.chain = chain
        self.max_window_size = max_window_size

    def forward_inverse_filling(self, input_data):
        """
        Method fills in the gaps in the input array using forward and inverse
        directions of predictions

        :param input_data: data with gaps to filling in the gaps in it
        :return: array without gaps
        """

        output_data = np.array(input_data)

        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iterately fill in the gaps in the time series
        for batch_index in range(len(new_gap_list)):

            preds = []
            weights = []
            # Two predictions are generated for each gap - forward and backward
            for direction_function in [self._forward, self._inverse]:

                weights_list, predicted_list = direction_function(output_data,
                                                                  batch_index,
                                                                  new_gap_list)
                weights.append(weights_list)
                preds.append(predicted_list)

            preds = np.array(preds)
            weights = np.array(weights)
            result = np.average(preds, axis=0, weights=weights)

            gap = new_gap_list[batch_index]
            # Replace gaps in an array with predicted values
            output_data[gap] = result

        return output_data

    def forward_filling(self, input_data):
        """
        Method fills in the gaps in the input array using chain with only
        forward direction (i.e. time series forecasting)

        :param input_data: data with gaps to filling in the gaps in it
        :return: array without gaps
        """

        output_data = np.array(input_data)

        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iterately fill in the gaps in the time series
        for gap in new_gap_list:
            # The entire time series is used for training until the gap
            timeseries_train_part = output_data[:gap[0]]

            # Adaptive prediction interval length
            len_gap = len(gap)

            # Chain for the task of filling in gaps
            predicted = self.__chain_fit_predict(timeseries_train_part,
                                                 len_gap)

            # Replace gaps in an array with predicted values
            output_data[gap] = predicted
        return output_data

    def _forward(self, timeseries_data, batch_index, new_gap_list):
        """
        The time series method makes a forward forecast based on the part
        of the time series that is located to the left of the gap.

        :param timeseries_data: one-dimensional array of a time series
        :param batch_index: index of the interval (batch) with a gap
        :param new_gap_list: array with nested lists of gap indexes

        :return weights_list: numpy array with prediction weights for
        averaging
        :return predicted_values: numpy array with predicted values in the
        gap
        """

        gap = new_gap_list[batch_index]
        timeseries_train_part = timeseries_data[:gap[0]]

        # Adaptive prediction interval length
        len_gap = len(gap)
        predicted_values = self.__chain_fit_predict(timeseries_train_part,
                                                    len_gap)
        weights_list = np.arange(len_gap, 0, -1)
        return weights_list, predicted_values

    def _inverse(self, timeseries_data, batch_index, new_gap_list):
        """
        The time series method makes an inverse forecast based on the part
        of the time series that is located to the right of the gap.

        :param timeseries_data: one-dimensional array of a time series
        :param batch_index: index of the interval (batch) with a gap
        :param new_gap_list: array with nested lists of gap indexes

        :return weights_list: numpy array with prediction weights for
        averaging
        :return predicted_values: numpy array with predicted values in the
        gap
        """

        gap = new_gap_list[batch_index]

        # If the interval with a gap is the last one in the array
        if batch_index == len(new_gap_list) - 1:
            timeseries_train_part = timeseries_data[(gap[-1] + 1):]
        else:
            next_gap = new_gap_list[batch_index + 1]
            timeseries_train_part = timeseries_data[(gap[-1] + 1):next_gap[0]]
        timeseries_train_part = np.flip(timeseries_train_part)

        # Adaptive prediction interval length
        len_gap = len(gap)

        predicted_values = self.__chain_fit_predict(timeseries_train_part,
                                                    len_gap)

        predicted_values = np.flip(predicted_values)
        weights_list = np.arange(1, (len_gap + 1), 1)
        return weights_list, predicted_values

    def __chain_fit_predict(self, timeseries_train: np.array, len_gap: int):
        """
        The method makes a prediction as a sequence of elements based on a
        training sample. There are two main parts: fit model and predict.

        :param timeseries_train: part of the time series for training the model
        :param len_gap: number of elements in the gap
        :return: array without gaps
        """

        task = Task(TaskTypesEnum.ts_forecasting,
                    TsForecastingParams(forecast_length=len_gap,
                                        max_window_size=self.max_window_size,
                                        return_all_steps=False,
                                        make_future_prediction=True))

        input_data = InputData(idx=np.arange(0, len(timeseries_train)),
                               features=None,
                               target=timeseries_train,
                               task=task,
                               data_type=DataTypesEnum.ts)

        # Making predictions for the missing part in the time series
        self.chain.fit_from_scratch(input_data)

        # "Test data" for making prediction for a specific length
        test_data = InputData(idx=np.arange(0, len_gap),
                              features=None,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.ts)

        predicted_values = self.chain.forecast(initial_data=input_data,
                                               supplementary_data=test_data).predict
        return predicted_values



import random
from typing import List

from fedot.core.composer.chain import Chain
from fedot.utilities.synthetic.chain_template import \
    (chain_template_balanced_tree, chain_template_random,
     fit_template, real_chain)


def chain_with_fixed_structure() -> Chain:
    """
    Generates chain with a fixed structure of nodes and links.
    :return:
    """
    raise NotImplementedError()


def chain_with_random_links(depth: int, models_per_level: List[int],
                            used_models: List[str]) -> Chain:
    """
    Generates chain with a fixed structure of nodes but random links.
    :param depth: Tree depth.
    :param models_per_level: The amount of models at each layer.
    :param used_models: The list of models to be randomly included into
    the resulted chain.
    :return: Chain with random links.
    """
    template = chain_template_random(model_types=used_models,
                                     depth=depth, models_per_level=models_per_level,
                                     samples=100, features=10)
    fit_template(chain_template=template, classes=2, skip_fit=True)
    resulted_chain = real_chain(chain_template=template)

    return resulted_chain


def chain_full_random(depth: int, max_level_size,
                      used_models: List[str]) -> Chain:
    """
    Generates chain with random amount of nodes and links.
    :param depth: Tree depth.
    :param max_level_size: Max amount of nodes per level.
    :param used_models: The list of models to be randomly included into
    the resulted chain.
    :return: Chain with random models_per_level and links.
    """
    models_per_lvl = _random_models_per_lvl(depth=depth,
                                            max_level_size=max_level_size)
    resulted_chain = chain_with_random_links(depth=depth,
                                             models_per_level=models_per_lvl,
                                             used_models=used_models)
    return resulted_chain


def chain_balanced_tree(depth: int, models_per_level: List[int],
                        used_models: List[str]) -> Chain:
    """
    Generates chain with balanced tree-like structure. The average arity value
    of the resulted tree is close to 2.
    :param depth: Tree depth.
    :param models_per_level: The amount of models at each layer.
    :param used_models: The list of models to be randomly included into
    the resulted chain.
    :return: Chain with balanced tree-like structure.
    """
    template = chain_template_balanced_tree(model_types=used_models,
                                            depth=depth,
                                            models_per_level=models_per_level,
                                            samples=100, features=10)
    fit_template(chain_template=template, classes=2, skip_fit=True)
    resulted_chain = real_chain(chain_template=template)

    return resulted_chain


def separately_fit_chain(samples: int, features_amount: int, classes: int,
                         chain: Chain = None):
    if chain is None:
        models = ['logit', 'xgboost',
                  'knn']
        template = chain_template_balanced_tree(model_types=models,
                                                depth=3,
                                                models_per_level=[4, 2, 1],
                                                samples=samples, features=features_amount)
        fit_template(chain_template=template, classes=classes, skip_fit=False)
        chain = real_chain(chain_template=template)

    return chain


def _random_models_per_lvl(depth, max_level_size):
    models_per_level = []
    for _ in range(depth - 1):
        models_per_level.append(random.randint(1, max_level_size))
    last_level_size = 1
    models_per_level.append(last_level_size)

    return models_per_level


if __name__ == '__main__':
    chain = chain_with_random_links(depth=3, models_per_level=[3, 2, 1],
                                    used_models=['xgboost', 'logit'])
    full_random_chain = chain_full_random(depth=4, max_level_size=4,
                                          used_models=['xgboost', 'logit'])



import gc
from typing import Optional

import h2o
import numpy as np
from h2o import H2OFrame
from h2o.automl import H2OAutoML
from tpot import TPOTClassifier, TPOTRegressor

from fedot.core.models.data import InputData
from fedot.core.models.evaluation.evaluation import EvaluationStrategy
from fedot.core.repository.tasks import TaskTypesEnum
from datetime import timedelta


def fit_tpot(data: InputData, max_run_time_min: int):
    models_hyperparameters = _get_models_hyperparameters(max_run_time_min)['TPOT']
    estimator = None
    if data.task.task_type == TaskTypesEnum.classification:
        estimator = TPOTClassifier
    elif data.task.task_type == TaskTypesEnum.regression:
        estimator = TPOTRegressor

    model = estimator(generations=models_hyperparameters['GENERATIONS'],
                      population_size=models_hyperparameters['POPULATION_SIZE'],
                      verbosity=2,
                      random_state=42,
                      max_time_mins=models_hyperparameters['MAX_RUNTIME_MINS'])

    model.fit(data.features, data.target)

    return model


def predict_tpot_reg(trained_model, predict_data):
    return trained_model.predict(predict_data.features)


def predict_tpot_class(trained_model, predict_data):
    try:
        return trained_model.predict_proba(predict_data.features)[:, 1]
    except AttributeError:
        # sklearn workaround for tpot
        return trained_model.predict(predict_data.features)


def fit_h2o(train_data: InputData, max_run_time_min: int):
    model_hyperparameters = _get_models_hyperparameters(max_run_time_min)['H2O']

    ip, port = _get_h2o_connect_config()

    h2o.init(ip=ip, port=port, name='h2o_server')

    frame = _data_transform(train_data)

    train_frame, valid_frame = frame.split_frame(ratios=[0.85])

    # make sure that your target column is the last one
    target_name = train_frame.columns[-1]
    predictor_names = train_frame.columns.remove(target_name)
    train_frame[target_name] = train_frame[target_name].asfactor()

    model = H2OAutoML(max_models=model_hyperparameters['MAX_MODELS'],
                      seed=1,
                      max_runtime_secs=model_hyperparameters['MAX_RUNTIME_SECS'])
    model.train(x=predictor_names, y=target_name, training_frame=train_frame, validation_frame=valid_frame)
    best_model = model.leader

    return best_model


def predict_h2o(trained_model, predict_data: InputData) -> np.array:
    test_frame = _data_transform(predict_data)

    target_name = test_frame.columns[-1]
    test_frame[target_name] = test_frame[target_name].asfactor()

    prediction_frame = trained_model.predict(test_frame)

    # return list of values like predict_proba[,:1] in sklearn
    prediction_proba_one: list = prediction_frame['p1'].transpose().getrow()

    return np.array(prediction_proba_one)


def _data_transform(data: InputData) -> H2OFrame:
    conc_data = np.concatenate((data.features, data.target.reshape(-1, 1)), 1)
    frame = H2OFrame(python_obj=conc_data)
    return frame


def _get_models_hyperparameters(timedelta: int = 5) -> dict:
    # MAX_RUNTIME_MINS should be equivalent to MAX_RUNTIME_SECS

    tpot_config = {'MAX_RUNTIME_MINS': timedelta,
                   'GENERATIONS': 50,
                   'POPULATION_SIZE': 10}

    h2o_config = {'MAX_MODELS': 20,
                  'MAX_RUNTIME_SECS': timedelta * 60}

    autokeras_config = {'MAX_TRIAL': 10,
                        'EPOCH': 100}

    space_for_mlbox = {
        'ne__numerical_strategy': {"space": [0, 'mean']},

        'ce__strategy': {"space": ["label_encoding", "random_projection", "entity_embedding"]},

        'fs__strategy': {"space": ["variance", "rf_feature_importance"]},
        'fs__threshold': {"search": "choice", "space": [0.1, 0.2, 0.3, 0.4, 0.5]},

        'est__strategy': {"space": ["LightGBM"]},
        'est__max_depth': {"search": "choice", "space": [5, 6]},
        'est__subsample': {"search": "uniform", "space": [0.6, 0.9]},
        'est__learning_rate': {"search": "choice", "space": [0.07]}

    }

    mlbox_config = {'space': space_for_mlbox, 'max_evals': 40}

    config_dictionary = {'TPOT': tpot_config, 'H2O': h2o_config,
                         'autokeras': autokeras_config, 'MLBox': mlbox_config}
    gc.collect()

    return config_dictionary


def _get_h2o_connect_config():
    IP = '127.0.0.1'
    PORT = 8888
    return IP, PORT


class AutoMLEvaluationStrategy(EvaluationStrategy):
    _model_functions_by_type = {
        'tpot': (fit_tpot, predict_tpot_class),
        'h2o': (fit_h2o, predict_h2o)
    }

    def __init__(self, model_type: str, params: Optional[dict] = None):
        self._model_specific_fit, self._model_specific_predict = \
            self._init_benchmark_model_functions(model_type)

        self.max_time_min = 5
        if params:
            self.max_time_min = params.get('max_run_time_sec', self.max_time_min * 60) / 60

        super().__init__(model_type, params)


    def _init_benchmark_model_functions(self, model_type):
        if model_type in self._model_functions_by_type.keys():
            return self._model_functions_by_type[model_type]
        else:
            raise ValueError(f'Impossible to obtain benchmark strategy for {model_type}')

    def fit(self, train_data: InputData):
        benchmark_model = self._model_specific_fit(train_data, self.max_time_min)
        return benchmark_model

    def predict(self, trained_model, predict_data: InputData):
        return self._model_specific_predict(trained_model, predict_data)

    def fit_tuned(self, train_data: InputData, iterations: int = 30,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        raise NotImplementedError()


class AutoMLRegressionStrategy(AutoMLEvaluationStrategy):
    _model_functions_by_type = {
        'tpot': (fit_tpot, predict_tpot_reg),
        'h2o': (fit_h2o, predict_h2o)
    }



from datetime import timedelta

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode
from fedot.core.log import default_log, Log, start_end_log_decorator
from fedot.core.models.data import InputData
from fedot.utilities.synthetic.chain_template_new import ChainTemplate, \
    ModelTemplate, extract_subtree_root


class Tune:

    def __init__(self, chain,
                 log: Log = default_log(__name__), verbose=False):
        self.chain = chain
        self.chain_template = ChainTemplate(self.chain)
        self.log = log
        self.verbose = verbose

    @start_end_log_decorator(start_msg='Starting tuning primary nodes',
                             end_msg='Primary nodes tuning is finished')
    def fine_tune_primary_nodes(self, input_data: InputData, iterations: int = 30,
                                max_lead_time: timedelta = timedelta(minutes=5)):
        """
        Optimize hyperparameters of models in primary nodes

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        :return: updated chain object
        """

        all_primary_nodes = [node for node in self.chain.nodes if isinstance(node, PrimaryNode)]
        for node in all_primary_nodes:
            node.fine_tune(input_data, max_lead_time=max_lead_time, iterations=iterations)

        return self.chain

    @start_end_log_decorator(start_msg='Starting tuning root node',
                             end_msg='Root node tuning is finished')
    def fine_tune_root_node(self, input_data: InputData, iterations: int = 30,
                            max_lead_time: timedelta = timedelta(minutes=5)):
        """
        Optimize hyperparameters in the root node

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        :return: updated chain object
        """

        node = self.chain.root_node
        node.fine_tune(input_data=input_data, max_lead_time=max_lead_time,
                       iterations=iterations, recursive=False)

        return self.chain

    @start_end_log_decorator(start_msg='Starting tuning all nodes',
                             end_msg='All nodes tuning is finished')
    def fine_tune_all_nodes(self, input_data: InputData, iterations: int = 30,
                            max_lead_time: timedelta = timedelta(minutes=5)):
        """
        Optimize hyperparameters of models in all nodes

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        :return: updated chain object
        """

        node = self.chain.root_node
        node.fine_tune(input_data, max_lead_time=max_lead_time, iterations=iterations, recursive=True)

        return self.chain

    @start_end_log_decorator(start_msg='Starting tuning chosen node',
                             end_msg='Chosen node tuning is finished')
    def fine_tune_certain_node(self, model_id, input_data: InputData, iterations: int = 30,
                               max_lead_time: timedelta = timedelta(minutes=5)):
        """
        Optimize hyperparameters of models in the certain node,
        defined by model id

        :param int model_id: number of the certain model in the chain.
        Look for it in exported json file of your model.
        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        :return: updated chain object
        """

        subchain = Chain()
        new_root = extract_subtree_root(root_model_id=model_id,
                                        chain_template=self.chain_template)
        subchain.add_node(new_root)
        subchain.fit(input_data=input_data, use_cache=False)

        updated_subchain = Tune(subchain).fine_tune_root_node(input_data=input_data,
                                                              iterations=iterations,
                                                              max_lead_time=max_lead_time)

        self._update_template(model_id=model_id,
                              updated_node=updated_subchain.root_node)

        updated_chain = Chain()
        self.chain_template.convert_to_chain(chain_to_convert_to=updated_chain)

        return updated_chain

    def _update_template(self, model_id, updated_node):
        model_template = [model_template for model_template in self.chain_template.model_templates
                          if model_template.model_id == model_id][0]
        update_node_template = ModelTemplate(updated_node, chain_id=self.chain_template.unique_chain_id)

        model_template.params = update_node_template.params
        model_template.fitted_model_path = update_node_template.fitted_model_path



from fedot.core.utils import ComparableEnum as Enum


class DataTypesEnum(Enum):
    table = 'feature_table'

    # 2d dataset with timeseries as target and external variables as features
    ts = 'time_series'

    # 2d dataset with time series forecasted by model
    forecasted_ts = 'time_series_forecasted'

    # 2d dataset with lagged features - (n, window_len * features)
    ts_lagged_table = 'time_series_lagged_table'



from copy import deepcopy

from hyperopt import tpe, fmin, Trials, space_eval, hp
from sklearn.model_selection import cross_val_score


def _search_space_transform(params):
    space = {}
    for param_name, param_space in params.items():
        space[param_name] = hp.choice(param_name, param_space)
    return space


class TunerAdapter:
    def __init__(self, tuner: 'Tuner'):
        self.tuner = tuner
        self.model_to_adapt = deepcopy(tuner.trained_model)
        self.data = tuner.tune_data
        self.params = _search_space_transform(tuner.params_range)
        self.scorer = tuner.scorer
        self.best_params: dict = dict()


class HyperoptAdapter(TunerAdapter):
    def __init__(self, tuner: 'TPE'):
        super(HyperoptAdapter, self).__init__(tuner)

    def _objective_function(self, params):
        self.model_to_adapt.set_params(**params)
        metric = cross_val_score(self.model_to_adapt,
                                 self.data.features,
                                 self.data.target,
                                 cv=5, scoring=self.scorer).mean()

        if self.greater_is_better():
            return -metric
        else:
            return metric

    def tune(self, iterations: int = 100, timeout_sec: int = 60):
        hyperopt_trials = Trials()
        best_encrypted = fmin(fn=self._objective_function,
                              space=self.params, algo=tpe.suggest,
                              max_evals=iterations,
                              trials=hyperopt_trials,
                              timeout=timeout_sec)
        self.best_params = space_eval(space=self.params,
                                      hp_assignment=best_encrypted)
        return self.best_params, self.best_model

    @property
    def best_model(self):
        return self.model_to_adapt.set_params(**self.best_params)

    def greater_is_better(self):
        """
        Extracting the private field of scorer to check
        whether the metric has a property of greater_is_better
        """
        is_greater = True if self.scorer._sign == 1 else False
        return is_greater



from datetime import timedelta

import numpy as np

from fedot.core.algorithms.time_series.prediction import post_process_forecasted_ts
from fedot.core.log import Log, default_log
from fedot.core.models.data import InputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.model_types_repository import ModelMetaInfo, ModelTypesRepository
from fedot.core.repository.tasks import Task, TaskTypesEnum, compatible_task_types

DEFAULT_PARAMS_STUB = 'default_params'


class Model:
    """
    Base object with fit/predict methods defining the evaluation strategy for the task

    :param model_type: str type of the model defined in model repository
    :param log: Log object to record messages
    """

    def __init__(self, model_type: str, log: Log = None):
        self.model_type = model_type
        self._eval_strategy, self._data_preprocessing = None, None
        self.params = DEFAULT_PARAMS_STUB
        self.log = log

        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

    @property
    def acceptable_task_types(self):
        model_info = ModelTypesRepository().model_info_by_id(self.model_type)
        return model_info.task_type

    @property
    def metadata(self) -> ModelMetaInfo:
        model_info = ModelTypesRepository().model_info_by_id(self.model_type)
        if not model_info:
            raise ValueError(f'Model {self.model_type} not found')
        return model_info

    def output_datatype(self, input_datatype: DataTypesEnum) -> DataTypesEnum:
        output_types = self.metadata.output_types
        if input_datatype in output_types:
            return input_datatype
        elif input_datatype == DataTypesEnum.ts:
            return DataTypesEnum.forecasted_ts
        else:
            return output_types[0]

    @property
    def description(self):
        model_type = self.model_type
        model_params = self.params
        return f'n_{model_type}_{model_params}'

    def _init(self, task: Task):
        params_for_fit = None
        if self.params != DEFAULT_PARAMS_STUB:
            params_for_fit = self.params

        try:
            self._eval_strategy = _eval_strategy_for_task(self.model_type, task.task_type)(self.model_type,
                                                                                           params_for_fit)
        except Exception as ex:
            self.log.error(f'Can not find evaluation strategy because of {ex}')
            raise ex

    def fit(self, data: InputData):
        """
        This method is used for defining and running of the evaluation strategy
        to train the model with the data provided

        :param data: data used for model training
        :return: tuple of trained model and prediction on train data
        """
        self._init(data.task)

        prepared_data = data.prepare_for_modelling(is_for_fit=True)

        fitted_model = self._eval_strategy.fit(train_data=prepared_data)

        predict_train = self.predict(fitted_model, data)

        return fitted_model, predict_train

    def predict(self, fitted_model, data: InputData):
        """
        This method is used for defining and running of the evaluation strategy
        to predict with the data provided

        :param fitted_model: trained model object
        :param data: data used for prediction
        """
        self._init(data.task)

        prepared_data = data.prepare_for_modelling(is_for_fit=False)

        prediction = self._eval_strategy.predict(trained_model=fitted_model,
                                                 predict_data=prepared_data)

        prediction = _post_process_prediction_using_original_input(prediction=prediction, input_data=data)

        return prediction

    def fine_tune(self, data: InputData, iterations: int,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        """
        This method is used for hyperparameter searching

        :param data: data used for hyperparameter searching
        :param iterations: max number of iterations evaluable for hyperparameter optimization
        :param max_lead_time: max time(seconds) for tuning evaluation
        """
        self._init(data.task)

        prepared_data = data.prepare_for_modelling(is_for_fit=True)

        try:
            fitted_model, tuned_params = self._eval_strategy.fit_tuned(train_data=prepared_data,
                                                                       iterations=iterations,
                                                                       max_lead_time=max_lead_time)
            self.params = tuned_params
            if not self.params:
                self.params = DEFAULT_PARAMS_STUB
        except Exception as ex:
            print(f'Tuning failed because of {ex}')
            fitted_model = self._eval_strategy.fit(train_data=data)
            self.params = DEFAULT_PARAMS_STUB

        predict_train = self.predict(fitted_model, data)

        return fitted_model, predict_train

    def __str__(self):
        return f'{self.model_type}'


def _eval_strategy_for_task(model_type: str, task_type_for_data: TaskTypesEnum):
    models_repo = ModelTypesRepository()
    model_info = models_repo.model_info_by_id(model_type)

    task_type_for_model = task_type_for_data
    task_types_acceptable_for_model = model_info.task_type

    # if the model can't be used directly for the task type from data
    if task_type_for_model not in task_types_acceptable_for_model:
        # search the supplementary task types, that can be included in chain which solves original task
        globally_compatible_task_types = compatible_task_types(task_type_for_model)
        compatible_task_types_acceptable_for_model = list(set(task_types_acceptable_for_model).intersection
                                                          (set(globally_compatible_task_types)))
        if len(compatible_task_types_acceptable_for_model) == 0:
            raise ValueError(f'Model {model_type} can not be used as a part of {task_type_for_model}.')
        task_type_for_model = compatible_task_types_acceptable_for_model[0]

    strategy = models_repo.model_info_by_id(model_type).current_strategy(task_type_for_model)
    return strategy


def _post_process_prediction_using_original_input(prediction, input_data: InputData):
    processed_predict = prediction
    if input_data.task.task_type == TaskTypesEnum.ts_forecasting:
        processed_predict = post_process_forecasted_ts(prediction, input_data)
    else:
        if np.array([np.isnan(_) for _ in prediction]).any():
            processed_predict = np.nan_to_num(prediction)

    return processed_predict



from copy import copy
import numpy as np

from fedot.core.composer.chain import Chain
from fedot.core.models.data import InputData, OutputData
from fedot.core.repository.tasks import TaskTypesEnum


class TsForecastingChain(Chain):

    def forecast(self, initial_data: InputData, supplementary_data: InputData) -> OutputData:
        """Generates the time series forecast with a sliding window using pre-fitted chain.
        :param initial_data: the initial condition for the forecasting (should be greater or equals to max_window_size)
        :param supplementary_data: the data that should be available during the forecast:
            idx for the forecasted steps and optional exogenous variables
            (variables that are received from an external source instead of forecasting in place and
            used as features of the forecasting model to increase the quality of of forecast)
        :return: forecasted time series
        """

        if not self.is_all_cache_actual():
            raise ValueError('Chain for the time series forecasting was not fitted yet.')

        if supplementary_data.task.task_type is not TaskTypesEnum.ts_forecasting:
            raise ValueError('TsForecastingChain can be used for the ts_forecasting task only.')

        supplementary_data_for_forecast = copy(supplementary_data)
        supplementary_data_for_forecast.task.task_params.make_future_prediction = True

        initial_data_for_forecast = copy(initial_data)
        initial_data_for_forecast.task.task_params.make_future_prediction = True

        forecast_length = supplementary_data_for_forecast.task.task_params.forecast_length

        # check if predict features contains additional (exogenous) variables
        with_exog = supplementary_data_for_forecast.features is not None

        # initial data for the first prediction
        pre_history_start = (len(initial_data_for_forecast.idx) -
                             initial_data_for_forecast.task.task_params.max_window_size)
        pre_history_end = len(initial_data_for_forecast.idx)

        data_for_forecast = initial_data_for_forecast.subset(start=pre_history_start, end=pre_history_end)

        full_prediction = []
        forecast_steps_num = int(np.ceil(len(supplementary_data_for_forecast.idx) / forecast_length))
        for forecast_step in range(forecast_steps_num):
            stepwise_prediction = self.predict(data_for_forecast).predict
            if len(stepwise_prediction.shape) > 1:
                # multi-dim prediction
                stepwise_prediction = stepwise_prediction[-1, :-forecast_length]
                full_prediction.extend(stepwise_prediction)
            else:
                # single-dim prediction
                stepwise_prediction = list(stepwise_prediction[-forecast_length:])
                full_prediction.extend(stepwise_prediction)

            # add additional variable from external source
            if with_exog:
                data_for_forecast = _prepare_exog_features(data_for_forecast, supplementary_data_for_forecast,
                                                           stepwise_prediction,
                                                           forecast_step, forecast_length)
            else:
                predicted_ts = np.append(data_for_forecast.target, stepwise_prediction)
                data_for_forecast.target = np.stack(predicted_ts)
                data_for_forecast.features = data_for_forecast.target

            data_for_forecast.idx = _extend_idx_for_prediction(data_for_forecast.idx, forecast_length)

        full_prediction = full_prediction[0:len(supplementary_data_for_forecast.idx)]

        output_data = OutputData(idx=supplementary_data_for_forecast.idx,
                                 features=supplementary_data_for_forecast.features,
                                 predict=np.asarray(full_prediction), task=supplementary_data_for_forecast.task,
                                 data_type=supplementary_data_for_forecast.data_type)

        return output_data


def _extend_idx_for_prediction(exiting_idx, forecast_length):
    if forecast_length > 1:
        indices_for_forecast = (exiting_idx[-1] +
                                list(range(1, forecast_length + 1)))

    else:
        indices_for_forecast = exiting_idx[-1] + 1

    new_idx = np.append(exiting_idx, indices_for_forecast)
    return new_idx


def _prepare_exog_features(data_for_prediction: InputData,
                           exog_data: InputData,
                           last_prediction: np.array,
                           forecast_step: int, forecast_length: int) -> InputData:
    new_features = []
    if len(data_for_prediction.features.shape) == 1:
        # if one exog feature
        exog_features_num = 1
    else:
        # if several exog features
        exog_features_num = data_for_prediction.features.shape[1]

    new_part_len = 0
    if exog_features_num > 1:
        for exog_feat_id in range(exog_features_num):
            exog_feature = data_for_prediction.features[:, exog_feat_id]
            new_exog_values = \
                exog_data.features[forecast_step * forecast_length:
                                   ((forecast_step + 1) * forecast_length), exog_feat_id]
            new_feature = np.append(exog_feature, new_exog_values)
            new_features.append(new_feature)
            new_part_len = len(new_features[0])
    else:
        exog_feature = data_for_prediction.features
        new_exog_values = \
            exog_data.features[forecast_step * forecast_length:
                               ((forecast_step + 1) * forecast_length)]
        new_features = np.append(exog_feature, new_exog_values)
        new_part_len = len(new_features)

    # add predicted time series to features for next prediction
    predicted_ts = np.append(data_for_prediction.target,
                             last_prediction)

    # cut the prediction if it's too long (actual for the last forecast step)
    predicted_ts = predicted_ts[0: new_part_len]
    # new_features.append(predicted_ts)
    data_for_prediction.target = predicted_ts
    data_for_prediction.features = np.stack(np.asarray(new_features)).T

    return data_for_prediction


def convert_to_tschain(chain):
    """
    The method convert base chain object into TsChain
    """

    chain.__class__ = TsForecastingChain
    return chain
